"""Abstract class to define the API for an SPH scheme.  The idea is that
one can define a scheme and thereafter one simply instantiates a suitable
scheme, gives it a bunch of particles and runs the application.
"""


class Scheme(object):

    """An API for an SPH scheme.
    """

    def __init__(self, fluids, solids, dim):
        """
        Parameters
        ----------

        fluids: list
            List of names of fluid particle arrays.
        solids: list
            List of names of solid particle arrays (or boundaries).
        dim: int
            Dimensionality of the problem.
        """
        self.fluids = fluids
        self.solids = solids
        self.dim = dim
        self.solver = None
        self.attributes_changed()

    # Public protocol ###################################################
    def add_user_options(self, group):
        pass

    def attributes_changed(self):
        """Overload this to compute any properties that depend on others.

        This is automatically called when configure is called.
        """
        pass

    def configure(self, **kw):
        """Configure the scheme with given parameters.

        Overload this to do any scheme specific stuff.
        """
        for k, v in kw.items():
            if not hasattr(self, k):
                msg = 'Parameter {param} not defined for {scheme}.'.format(
                    param=k, scheme=self.__class__.__name__
                )
                raise RuntimeError(msg)
            setattr(self, k, v)
        self.attributes_changed()

    def consume_user_options(self, options):
        pass

    def configure_solver(self, kernel=None, integrator_cls=None,
                         extra_steppers=None, **kw):
        """Configure the solver to be generated.

        Parameters
        ----------

        kernel : Kernel instance.
            Kernel to use, if none is passed a default one is used.
        integrator_cls : pysph.sph.integrator.Integrator
            Integrator class to use, use sensible default if none is
            passed.
        extra_steppers : dict
            Additional integration stepper instances as a dict.
        **kw : extra arguments
            Any additional keyword args are passed to the solver instance.
        """
        raise NotImplementedError()

    def get_equations(self):
        raise NotImplementedError()

    def get_solver(self):
        return self.solver

    def setup_properties(self, particles, clean=True):
        """Setup the particle arrays so they have the right set of properties
        for this scheme.

        Parameters
        ----------

        particles : list
            List of particle arrays.

        clean : bool
            If True, removes any unnecessary properties.
        """
        raise NotImplementedError()

    # Private protocol ###################################################

    def _ensure_properties(self, pa, desired_props, clean=True):
        """Given a particle array and a set of properties desired,
        this removes unnecessary properties (if `clean=True`), and
        adds the desired properties.

        Parameters
        ----------

        pa : ParticleArray
            Desired particle array.
        desired_props : sequence
            Desired properties to have in the array, can be a list of strings
            or dicts with stride info or both.
        clean : bool
            Remove undesirable properties.
        """
        all_props = {}
        for p in desired_props:
            if isinstance(p, dict):
                all_props.update({p['name']: p})
            elif p not in all_props:
                all_props.update({p: {'name': p}})

        pa_props = set(pa.properties.keys())
        if clean:
            to_remove = pa_props - set(all_props.keys())
            for prop in to_remove:
                pa.remove_property(prop)

        to_add = set(all_props.keys()) - pa_props
        for prop in to_add:
            pa.add_property(**all_props[prop])

    def _smart_getattr(self, obj, var):
        res = getattr(obj, var)
        if res is None:
            return getattr(self, var)
        else:
            return res


class SchemeChooser(Scheme):
    def __init__(self, default, **schemes):
        """
        Parameters
        ----------

        default: str
            Name of the default scheme to use.
        **schemes: kwargs
            The schemes to choose between.
        """
        self.default = default
        self.schemes = dict(schemes)
        self.scheme = schemes[default]

    def add_user_options(self, group):
        for scheme in self.schemes.values():
            scheme.add_user_options(group)
        choices = list(self.schemes.keys())
        group.add_argument(
            "--scheme", action="store", dest="scheme",
            default=self.default, choices=choices,
            help="Specify scheme to use (one of %s)." % choices
        )

    def attributes_changed(self):
        self.scheme.attributes_changed()

    def configure(self, **kw):
        self.scheme.configure(**kw)

    def consume_user_options(self, options):
        self.scheme = self.schemes[options.scheme]
        self.scheme.consume_user_options(options)

    def configure_solver(self, kernel=None, integrator_cls=None,
                         extra_steppers=None, **kw):
        self.scheme.configure_solver(
            kernel=kernel, integrator_cls=integrator_cls,
            extra_steppers=extra_steppers, **kw
        )

    def get_equations(self):
        return self.scheme.get_equations()

    def get_solver(self):
        return self.scheme.get_solver()

    def setup_properties(self, particles, clean=True):
        """Setup the particle arrays so they have the right set of properties
        for this scheme.

        Parameters
        ----------

        particles : list
            List of particle arrays.

        clean : bool
            If True, removes any unnecessary properties.
        """
        self.scheme.setup_properties(particles, clean)


############################################################################

def add_bool_argument(group, arg, dest, help, default):
    group.add_argument(
        '--%s' % arg, action="store_true", dest=dest, help=help
    )
    neg_help = 'Do not ' + help[0].lower() + help[1:]
    group.add_argument(
        '--no-%s' % arg, action="store_false", dest=dest, help=neg_help
    )
    group.set_defaults(**{dest: default})


class WCSPHScheme(Scheme):
    def __init__(self, fluids, solids, dim, rho0, c0, h0, hdx, gamma=7.0,
                 gx=0.0, gy=0.0, gz=0.0, alpha=0.1, beta=0.0, delta=0.1,
                 nu=0.0, tensile_correction=False, hg_correction=False,
                 update_h=False, delta_sph=False, summation_density=False):
        """Parameters
        ----------

        fluids: list
            List of names of fluid particle arrays.
        solids: list
            List of names of solid particle arrays (or boundaries).
        dim: int
            Dimensionality of the problem.
        rho0: float
            Reference density.
        c0: float
            Reference speed of sound.
        gamma: float
            Gamma for the equation of state.
        h0: float
            Reference smoothing length.
        hdx: float
            Ratio of h/dx.
        gx, gy, gz: float
            Body force acceleration components.
        alpha: float
            Coefficient for artificial viscosity.
        beta: float
            Coefficient for artificial viscosity.
        delta: float
            Coefficient used to control the intensity of diffusion of density
        nu: float
            Real viscosity of the fluid, defaults to no viscosity.
        tensile_correction: bool
            Use tensile correction.
        hg_correction: bool
            Use the Hughes-Graham correction.
        update_h: bool
            Update the smoothing length as per Ferrari et al.
        delta_sph: bool
            Use the delta-SPH correction terms.
        summation_density: bool
            Use summation density instead of continuity.

        References
        ----------

        .. [Hughes2010] J. P. Hughes and D. I. Graham, "Comparison of
           incompressible and weakly-compressible SPH models for free-surface
           water flows", Journal of Hydraulic Research, 48 (2010), pp. 105-117.

        .. [Marrone2011] S. Marrone et al., "delta-SPH model for simulating
           violent impact flows", Computer Methods in Applied Mechanics and
           Engineering, 200 (2011), pp 1526--1542.

        .. [Cherfils2012] J. M. Cherfils et al., "JOSEPHINE: A parallel SPH
           code for free-surface flows", Computer Physics Communications, 183
           (2012), pp 1468--1480.

        """
        self.fluids = fluids
        self.solids = solids
        self.solver = None
        self.rho0 = rho0
        self.c0 = c0
        self.gamma = gamma
        self.dim = dim
        self.h0 = h0
        self.hdx = hdx
        self.gx = gx
        self.gy = gy
        self.gz = gz
        self.alpha = alpha
        self.beta = beta
        self.delta = delta
        self.nu = nu
        self.tensile_correction = tensile_correction
        self.hg_correction = hg_correction
        self.update_h = update_h
        self.delta_sph = delta_sph
        self.summation_density = summation_density

    def add_user_options(self, group):
        group.add_argument(
            "--alpha", action="store", type=float, dest="alpha",
            default=None,
            help="Alpha for the artificial viscosity."
        )
        group.add_argument(
            "--beta", action="store", type=float, dest="beta",
            default=None,
            help="Beta for the artificial viscosity."
        )
        group.add_argument(
            "--delta", action="store", type=float, dest="delta",
            default=None,
            help="Delta for the delta-SPH."
        )
        group.add_argument(
            "--gamma", action="store", type=float, dest="gamma",
            default=None,
            help="Gamma for the state equation."
        )
        add_bool_argument(
            group, 'tensile-correction', dest='tensile_correction',
            help="Use tensile instability correction.",
            default=None
        )
        add_bool_argument(
            group, "hg-correction", dest="hg_correction",
            help="Use the Hughes Graham correction.",
            default=None
        )
        add_bool_argument(
            group, "update-h", dest="update_h",
            help="Update the smoothing length as per Ferrari et al.",
            default=None
        )
        add_bool_argument(
            group, "delta-sph", dest="delta_sph",
            help="Use the delta-SPH correction terms.",
            default=None
        )
        add_bool_argument(
            group, "summation-density", dest="summation_density",
            help="Use summation density instead of continuity.",
            default=None
        )

    def consume_user_options(self, options):
        vars = ['gamma', 'tensile_correction', 'hg_correction',
                'update_h', 'delta_sph', 'alpha', 'beta',
                'summation_density']

        data = dict((var, self._smart_getattr(options, var))
                    for var in vars)
        self.configure(**data)

    def get_timestep(self, cfl=0.5):
        return cfl*self.h0/self.c0

    def configure_solver(self, kernel=None, integrator_cls=None,
                         extra_steppers=None, **kw):
        from pysph.base.kernels import CubicSpline
        if kernel is None:
            kernel = CubicSpline(dim=self.dim)

        steppers = {}
        if extra_steppers is not None:
            steppers.update(extra_steppers)

        from pysph.sph.integrator import PECIntegrator, TVDRK3Integrator
        from pysph.sph.integrator_step import WCSPHStep, WCSPHTVDRK3Step

        cls = integrator_cls if integrator_cls is not None else PECIntegrator
        step_cls = WCSPHTVDRK3Step if cls is TVDRK3Integrator else WCSPHStep
        for name in self.fluids + self.solids:
            if name not in steppers:
                steppers[name] = step_cls()

        integrator = cls(**steppers)

        from pysph.solver.solver import Solver
        if 'dt' not in kw:
            kw['dt'] = self.get_timestep()
        self.solver = Solver(
            dim=self.dim, integrator=integrator, kernel=kernel, **kw
        )

    def get_equations(self):
        from pysph.sph.equation import Group
        from pysph.sph.wc.basic import (
            MomentumEquation, TaitEOS, TaitEOSHGCorrection,
            UpdateSmoothingLengthFerrari
        )
        from pysph.sph.wc.basic import (ContinuityEquationDeltaSPH,
                                        ContinuityEquationDeltaSPHPreStep,
                                        MomentumEquationDeltaSPH)
        from pysph.sph.basic_equations import \
            (ContinuityEquation, SummationDensity, XSPHCorrection)
        from pysph.sph.wc.viscosity import (LaminarViscosity,
                                            LaminarViscosityDeltaSPH)
        from pysph.sph.wc.kernel_correction import (GradientCorrectionPreStep,
                                                    GradientCorrection)

        equations = []
        g1 = []
        all = self.fluids + self.solids

        if self.summation_density:
            g0 = []
            for name in self.fluids:
                g0.append(SummationDensity(dest=name, sources=all))
            equations.append(Group(equations=g0, real=False))

        for name in self.fluids:
            g1.append(TaitEOS(
                dest=name, sources=None, rho0=self.rho0, c0=self.c0,
                gamma=self.gamma
            ))

        # This correction applies only to solids.
        for name in self.solids:
            if self.hg_correction:
                g1.append(TaitEOSHGCorrection(
                    dest=name, sources=None, rho0=self.rho0, c0=self.c0,
                    gamma=self.gamma
                ))
            else:
                g1.append(TaitEOS(
                    dest=name, sources=None, rho0=self.rho0, c0=self.c0,
                    gamma=self.gamma
                ))
        equations.append(Group(equations=g1, real=False))

        if self.delta_sph and not self.summation_density:
            eq2_pre = []
            for name in self.fluids:
                eq2_pre.append(
                    GradientCorrectionPreStep(dest=name, sources=[name],
                                              dim=self.dim)
                )
            equations.append(Group(equations=eq2_pre, real=False))

            eq2 = []
            for name in self.fluids:
                eq2.extend([
                    GradientCorrection(dest=name, sources=[name]),
                    ContinuityEquationDeltaSPHPreStep(
                        dest=name, sources=[name]
                    )])
            equations.append(Group(equations=eq2))

        g2 = []
        for name in self.solids:
            g2.append(ContinuityEquation(dest=name, sources=self.fluids))

        for name in self.fluids:
            if not self.summation_density:
                g2.append(
                    ContinuityEquation(dest=name, sources=all)
                )
            if self.delta_sph and not self.summation_density:
                g2.append(
                    ContinuityEquationDeltaSPH(
                        dest=name, sources=[name], c0=self.c0,
                        delta=self.delta
                    ))
            # This is required since MomentumEquation (ME) adds artificial
            # viscosity (AV), so make alpha 0.0 for ME and enable delta sph AV.
            alpha = 0.0 if self.delta_sph else self.alpha
            g2.append(
                MomentumEquation(
                    dest=name, sources=all, c0=self.c0,
                    alpha=alpha, beta=self.beta,
                    gx=self.gx, gy=self.gy, gz=self.gz,
                    tensile_correction=self.tensile_correction
                ))
            if self.delta_sph:
                g2.append(
                    MomentumEquationDeltaSPH(
                        dest=name, sources=[name], rho0=self.rho0, c0=self.c0,
                        alpha=self.alpha
                    ))
            g2.append(XSPHCorrection(dest=name, sources=[name]))

            if abs(self.nu) > 1e-14:
                if self.delta_sph:
                    eq = LaminarViscosityDeltaSPH(
                        dest=name, sources=all, dim=self.dim, rho0=self.rho0,
                        nu=self.nu
                    )
                else:
                    eq = LaminarViscosity(
                        dest=name, sources=all, nu=self.nu
                    )
                g2.insert(-1, eq)
        equations.append(Group(equations=g2))

        if self.update_h:
            g3 = [
                UpdateSmoothingLengthFerrari(
                    dest=x, sources=None, dim=self.dim, hdx=self.hdx
                ) for x in self.fluids
            ]
            equations.append(Group(equations=g3, real=False))

        return equations

    def setup_properties(self, particles, clean=True):
        from pysph.base.utils import get_particle_array_wcsph
        dummy = get_particle_array_wcsph(name='junk')
        props = list(dummy.properties.keys())
        output_props = ['x', 'y', 'z', 'u', 'v', 'w', 'rho', 'm', 'h',
                        'pid', 'gid', 'tag', 'p']
        if self.delta_sph:
            delta_sph_props = [
                {'name': 'm_mat', 'stride': 9},
                {'name': 'gradrho', 'stride': 3},
            ]
            props += delta_sph_props
        for pa in particles:
            self._ensure_properties(pa, props, clean)
            pa.set_output_arrays(output_props)
            if pa.name in self.solids:
                # This is the load balancing weight for the solid particles.
                # They do less work so we reduce the weight.
                if 'lb_weight' not in pa.constants:
                    pa.add_constant('lb_weight', 0.1)


class TVFScheme(Scheme):
    def __init__(self, fluids, solids, dim, rho0, c0, nu, p0, pb, h0,
                 gx=0.0, gy=0.0, gz=0.0, alpha=0.0, tdamp=0.0):
        self.fluids = fluids
        self.solids = solids
        self.solver = None
        self.rho0 = rho0
        self.c0 = c0
        self.pb = pb
        self.p0 = p0
        self.nu = nu
        self.dim = dim
        self.h0 = h0
        self.gx = gx
        self.gy = gy
        self.gz = gz
        self.alpha = alpha
        self.tdamp = 0.0

    def add_user_options(self, group):
        group.add_argument(
            "--alpha", action="store", type=float, dest="alpha",
            default=None,
            help="Alpha for the artificial viscosity."
        )
        group.add_argument(
            "--tdamp", action="store", type=float, dest="tdamp",
            default=None,
            help="Time for which the accelerations are damped."
        )

    def consume_user_options(self, options):
        vars = ['alpha', 'tdamp']
        data = dict((var, self._smart_getattr(options, var))
                    for var in vars)
        self.configure(**data)

    def get_timestep(self, cfl=0.25):
        dt_cfl = cfl * self.h0/self.c0
        if self.nu > 1e-12:
            dt_viscous = 0.125 * self.h0**2/self.nu
        else:
            dt_viscous = 1.0
        dt_force = 1.0

        return min(dt_cfl, dt_viscous, dt_force)

    def configure_solver(self, kernel=None, integrator_cls=None,
                         extra_steppers=None, **kw):
        """Configure the solver to be generated.

        Parameters
        ----------

        kernel : Kernel instance.
            Kernel to use, if none is passed a default one is used.
        integrator_cls : pysph.sph.integrator.Integrator
            Integrator class to use, use sensible default if none is
            passed.
        extra_steppers : dict
            Additional integration stepper instances as a dict.
        **kw : extra arguments
            Any additional keyword args are passed to the solver instance.
        """
        from pysph.base.kernels import QuinticSpline
        from pysph.sph.integrator_step import TransportVelocityStep
        from pysph.sph.integrator import PECIntegrator
        if kernel is None:
            kernel = QuinticSpline(dim=self.dim)
        steppers = {}
        if extra_steppers is not None:
            steppers.update(extra_steppers)

        step_cls = TransportVelocityStep
        for fluid in self.fluids:
            if fluid not in steppers:
                steppers[fluid] = step_cls()

        cls = integrator_cls if integrator_cls is not None else PECIntegrator
        integrator = cls(**steppers)

        from pysph.solver.solver import Solver
        self.solver = Solver(
            dim=self.dim, integrator=integrator, kernel=kernel, **kw
        )

    def get_equations(self):
        from pysph.sph.equation import Group
        from pysph.sph.wc.transport_velocity import (
            SummationDensity, StateEquation, MomentumEquationPressureGradient,
            MomentumEquationArtificialViscosity,
            MomentumEquationViscosity, MomentumEquationArtificialStress,
            SolidWallPressureBC, SolidWallNoSlipBC, SetWallVelocity
        )
        equations = []
        all = self.fluids + self.solids
        g1 = []
        for fluid in self.fluids:
            g1.append(SummationDensity(dest=fluid, sources=all))

        equations.append(Group(equations=g1, real=False))

        g2 = []
        for fluid in self.fluids:
            g2.append(StateEquation(
                dest=fluid, sources=None, p0=self.p0, rho0=self.rho0, b=1.0
            ))
        for solid in self.solids:
            g2.append(SetWallVelocity(dest=solid, sources=self.fluids))

        if len(g2) > 0:
            equations.append(Group(equations=g2, real=False))

        g3 = []
        for solid in self.solids:
            g3.append(SolidWallPressureBC(
                dest=solid, sources=self.fluids, b=1.0, rho0=self.rho0,
                p0=self.p0, gx=self.gx, gy=self.gy, gz=self.gz
            ))

        if len(g3) > 0:
            equations.append(Group(equations=g3, real=False))

        g4 = []
        for fluid in self.fluids:
            g4.append(
                MomentumEquationPressureGradient(
                    dest=fluid, sources=all, pb=self.pb, gx=self.gx,
                    gy=self.gy, gz=self.gz, tdamp=self.tdamp
                )
            )
            if self.alpha > 0.0:
                g4.append(
                    MomentumEquationArtificialViscosity(
                        dest=fluid, sources=all, c0=self.c0,
                        alpha=self.alpha
                    )
                )
            if self.nu > 0.0:
                g4.append(
                    MomentumEquationViscosity(
                        dest=fluid, sources=self.fluids, nu=self.nu
                    )
                )
                if len(self.solids) > 0:
                    g4.append(
                        SolidWallNoSlipBC(
                            dest=fluid, sources=self.solids, nu=self.nu
                        )
                    )

            g4.append(
                MomentumEquationArtificialStress(
                    dest=fluid, sources=self.fluids)
            )

        equations.append(Group(equations=g4))
        return equations

    def setup_properties(self, particles, clean=True):
        from pysph.base.utils import get_particle_array_tvf_fluid, \
            get_particle_array_tvf_solid
        particle_arrays = dict([(p.name, p) for p in particles])
        dummy = get_particle_array_tvf_fluid(name='junk')
        props = list(dummy.properties.keys())
        output_props = dummy.output_property_arrays
        for fluid in self.fluids:
            pa = particle_arrays[fluid]
            self._ensure_properties(pa, props, clean)
            pa.set_output_arrays(output_props)

        dummy = get_particle_array_tvf_solid(name='junk')
        props = list(dummy.properties.keys())
        output_props = dummy.output_property_arrays
        for solid in self.solids:
            pa = particle_arrays[solid]
            self._ensure_properties(pa, props, clean)
            pa.set_output_arrays(output_props)


class AdamiHuAdamsScheme(TVFScheme):
    """This is a scheme similiar to that in the paper:

    Adami, S., Hu, X., Adams, N. A generalized wall boundary condition for
    smoothed particle hydrodynamics.  Journal of Computational Physics
    2012;231(21):7057-7075.

    The major difference is in how the equations are integrated.  The paper
    has a different scheme that does not quite fit in with how things are done
    in PySPH readily so we simply use the WCSPHStep which works well.
    """
    def __init__(self, fluids, solids, dim, rho0, c0, nu, h0,
                 gx=0.0, gy=0.0, gz=0.0, p0=0.0, gamma=7.0,
                 tdamp=0.0, alpha=0.0):
        self.fluids = fluids
        self.solids = solids
        self.solver = None
        self.rho0 = rho0
        self.c0 = c0
        self.h0 = h0
        self.p0 = p0
        self.nu = nu
        self.dim = dim
        self.gx = gx
        self.gy = gy
        self.gz = gz
        self.alpha = alpha
        self.gamma = float(gamma)
        self.tdamp = tdamp
        self.attributes_changed()

    def add_user_options(self, group):
        super(AdamiHuAdamsScheme, self).add_user_options(group)
        group.add_argument(
            "--gamma", action="store", type=float, dest="gamma",
            default=None,
            help="Gamma for the state equation."
        )

    def attributes_changed(self):
        self.B = self.c0*self.c0*self.rho0/self.gamma

    def consume_user_options(self, options):
        vars = ['alpha', 'tdamp', 'gamma']
        data = dict((var, self._smart_getattr(options, var))
                    for var in vars)
        self.configure(**data)

    def configure_solver(self, kernel=None, integrator_cls=None,
                         extra_steppers=None, **kw):
        """Configure the solver to be generated.

        Parameters
        ----------

        kernel : Kernel instance.
            Kernel to use, if none is passed a default one is used.
        integrator_cls : pysph.sph.integrator.Integrator
            Integrator class to use, use sensible default if none is
            passed.
        extra_steppers : dict
            Additional integration stepper instances as a dict.
        **kw : extra arguments
            Any additional keyword args are passed to the solver instance.
        """
        from pysph.base.kernels import QuinticSpline
        from pysph.sph.integrator_step import WCSPHStep
        from pysph.sph.integrator import PECIntegrator
        if kernel is None:
            kernel = QuinticSpline(dim=self.dim)
        steppers = {}
        if extra_steppers is not None:
            steppers.update(extra_steppers)

        step_cls = WCSPHStep
        for fluid in self.fluids:
            if fluid not in steppers:
                steppers[fluid] = step_cls()

        cls = integrator_cls if integrator_cls is not None else PECIntegrator
        integrator = cls(**steppers)

        from pysph.solver.solver import Solver
        self.solver = Solver(
            dim=self.dim, integrator=integrator, kernel=kernel, **kw
        )

    def get_equations(self):
        from pysph.sph.equation import Group
        from pysph.sph.wc.basic import TaitEOS
        from pysph.sph.basic_equations import XSPHCorrection
        from pysph.sph.wc.transport_velocity import (
            ContinuityEquation, ContinuitySolid,
            MomentumEquationPressureGradient,
            MomentumEquationViscosity, MomentumEquationArtificialViscosity,
            SolidWallPressureBC, SolidWallNoSlipBC, SetWallVelocity,
            VolumeSummation
        )

        equations = []
        all = self.fluids + self.solids

        g2 = []
        for fluid in self.fluids:
            g2.append(VolumeSummation(dest=fluid, sources=all))
            g2.append(TaitEOS(
                dest=fluid, sources=None, rho0=self.rho0, c0=self.c0,
                gamma=self.gamma, p0=self.p0
            ))
        for solid in self.solids:
            g2.append(VolumeSummation(dest=solid, sources=all))
            g2.append(SetWallVelocity(dest=solid, sources=self.fluids))

        equations.append(Group(equations=g2, real=False))

        g3 = []
        for solid in self.solids:
            g3.append(SolidWallPressureBC(
                dest=solid, sources=self.fluids, b=1.0, rho0=self.rho0,
                p0=self.B, gx=self.gx, gy=self.gy, gz=self.gz
            ))

        equations.append(Group(equations=g3, real=False))

        g4 = []
        for fluid in self.fluids:
            g4.append(
                ContinuityEquation(dest=fluid, sources=self.fluids)
            )
            if self.solids:
                g4.append(
                    ContinuitySolid(dest=fluid, sources=self.solids)
                )
            g4.append(
                MomentumEquationPressureGradient(
                    dest=fluid, sources=all, pb=0.0, gx=self.gx,
                    gy=self.gy, gz=self.gz, tdamp=self.tdamp
                )
            )
            if self.alpha > 0.0:
                g4.append(
                    MomentumEquationArtificialViscosity(
                        dest=fluid, sources=all, c0=self.c0,
                        alpha=self.alpha
                    )
                )
            if self.nu > 0.0:
                g4.append(
                    MomentumEquationViscosity(
                        dest=fluid, sources=self.fluids, nu=self.nu
                    )
                )
                if len(self.solids) > 0:
                    g4.append(
                        SolidWallNoSlipBC(
                            dest=fluid, sources=self.solids, nu=self.nu
                        )
                    )
            g4.append(XSPHCorrection(dest=fluid, sources=[fluid]))

        equations.append(Group(equations=g4))
        return equations

    def setup_properties(self, particles, clean=True):
        super(AdamiHuAdamsScheme, self).setup_properties(particles, clean)
        particle_arrays = dict([(p.name, p) for p in particles])
        props = ['cs', 'arho', 'rho0', 'u0', 'v0', 'w0', 'x0', 'y0', 'z0',
                 'ax', 'ay', 'az']
        for fluid in self.fluids:
            pa = particle_arrays[fluid]
            for prop in props:
                pa.add_property(prop)


class GasDScheme(Scheme):
    def __init__(self, fluids, solids, dim, gamma, kernel_factor, alpha1=1.0,
                 alpha2=0.1, beta=2.0, adaptive_h_scheme='mpm',
                 update_alpha1=False, update_alpha2=False,
                 max_density_iterations=250,
                 density_iteration_tolerance=1e-3, has_ghosts=False):
        """
        Parameters
        ----------

        fluids: list
            List of names of fluid particle arrays.
        solids: list
            List of names of solid particle arrays (or boundaries), currently
            not supported
        dim: int
            Dimensionality of the problem.
        gamma: float
            Gamma for Equation of state.
        kernel_factor: float
            Kernel scaling factor.
        alpha1: float
            Artificial viscosity parameter.
        alpha2: float
            Artificial viscosity parameter.
        beta: float
            Artificial viscosity parameter.
        adaptive_h_scheme: str
            Adaptive h scheme to use. One of ['mpm', 'gsph']
        update_alpha1: bool
            Update the alpha1 parameter dynamically.
        update_alpha2: bool
            Update the alpha2 parameter dynamically.
        max_density_iterations: int
            Maximum number of iterations to run for one density step
        density_iteration_tolerance: float
            Maximum difference allowed in two successive density iterations
        has_ghosts: bool
            if ghost particles (either mirror or periodic) is used
        """
        self.fluids = fluids
        self.solids = solids
        self.dim = dim
        self.solver = None
        self.gamma = gamma
        self.alpha1 = alpha1
        self.alpha2 = alpha2
        self.update_alpha1 = update_alpha1
        self.update_alpha2 = update_alpha2
        self.beta = beta
        self.kernel_factor = kernel_factor
        self.adaptive_h_scheme = adaptive_h_scheme
        self.density_iteration_tolerance = density_iteration_tolerance
        self.max_density_iterations = max_density_iterations
        self.has_ghosts = has_ghosts

    def add_user_options(self, group):
        choices = ['gsph', 'mpm']
        group.add_argument(
            "--adaptive-h", action="store", dest="adaptive_h_scheme",
            default=None, choices=choices,
            help="Specify scheme for adaptive smoothing lengths %s" % choices
        )
        group.add_argument(
            "--alpha1", action="store", type=float, dest="alpha1",
            default=None,
            help="Alpha1 for the artificial viscosity."
        )
        group.add_argument(
            "--beta", action="store", type=float, dest="beta",
            default=None,
            help="Beta for the artificial viscosity."
        )
        group.add_argument(
            "--alpha2", action="store", type=float, dest="alpha2",
            default=None,
            help="Alpha2 for artificial viscosity"
        )
        group.add_argument(
            "--gamma", action="store", type=float, dest="gamma",
            default=None,
            help="Gamma for the state equation."
        )
        add_bool_argument(
            group, "update-alpha1", dest="update_alpha1",
            help="Update the alpha1 parameter.",
            default=None
        )
        add_bool_argument(
            group, "update-alpha2", dest="update_alpha2",
            help="Update the alpha2 parameter.",
            default=None
        )

    def consume_user_options(self, options):
        vars = ['gamma', 'alpha2', 'alpha1', 'beta', 'update_alpha1',
                'update_alpha2', 'adaptive_h_scheme']
        data = dict((var, self._smart_getattr(options, var))
                    for var in vars)
        self.configure(**data)

    def configure_solver(self, kernel=None, integrator_cls=None,
                         extra_steppers=None, **kw):
        """Configure the solver to be generated.

        Parameters
        ----------

        kernel : Kernel instance.
            Kernel to use, if none is passed a default one is used.
        integrator_cls : pysph.sph.integrator.Integrator
            Integrator class to use, use sensible default if none is
            passed.
        extra_steppers : dict
            Additional integration stepper instances as a dict.
        **kw : extra arguments
            Any additional keyword args are passed to the solver instance.
        """
        from pysph.base.kernels import Gaussian
        if kernel is None:
            kernel = Gaussian(dim=self.dim)

        steppers = {}
        if extra_steppers is not None:
            steppers.update(extra_steppers)

        from pysph.sph.integrator import PECIntegrator
        from pysph.sph.integrator_step import GasDFluidStep

        cls = integrator_cls if integrator_cls is not None else PECIntegrator
        step_cls = GasDFluidStep
        for name in self.fluids:
            if name not in steppers:
                steppers[name] = step_cls()

        integrator = cls(**steppers)

        from pysph.solver.solver import Solver
        self.solver = Solver(
            dim=self.dim, integrator=integrator, kernel=kernel, **kw
        )

    def get_equations(self):
        from pysph.sph.equation import Group
        from pysph.sph.gas_dynamics.basic import (
            ScaleSmoothingLength, UpdateSmoothingLengthFromVolume,
            SummationDensity, IdealGasEOS, MPMAccelerations,
            MPMUpdateGhostProps
        )
        from pysph.sph.gas_dynamics.boundary_equations import WallBoundary

        equations = []
        # Find the optimal 'h'
        if self.adaptive_h_scheme == 'mpm':
            g1 = []
            for fluid in self.fluids:
                g1.append(
                    SummationDensity(
                        dest=fluid, sources=self.fluids, k=self.kernel_factor,
                        density_iterations=True, dim=self.dim,
                        htol=self.density_iteration_tolerance
                    )
                )

            equations.append(Group(
                equations=g1, update_nnps=True, iterate=True,
                max_iterations=self.max_density_iterations
            ))

        elif self.adaptive_h_scheme == 'gsph':
            group = []
            for fluid in self.fluids:
                group.append(
                    ScaleSmoothingLength(dest=fluid, sources=None, factor=2.0)
                )
            equations.append(Group(equations=group, update_nnps=True))

            group = []
            for fluid in self.fluids:
                group.append(
                    SummationDensity(
                        dest=fluid, sources=self.fluids, dim=self.dim
                    )
                )
            equations.append(Group(equations=group, update_nnps=False))

            group = []
            for fluid in self.fluids:
                group.append(
                    UpdateSmoothingLengthFromVolume(
                        dest=fluid, sources=None, k=self.kernel_factor,
                        dim=self.dim
                    )
                )
            equations.append(Group(equations=group, update_nnps=True))

            group = []
            for fluid in self.fluids:
                group.append(
                    SummationDensity(
                        dest=fluid, sources=self.fluids, dim=self.dim
                    )
                )
            equations.append(Group(equations=group, update_nnps=False))
        # Done with finding the optimal 'h'

        g2 = []
        for fluid in self.fluids:
            g2.append(IdealGasEOS(dest=fluid, sources=None, gamma=self.gamma))

        equations.append(Group(equations=g2))

        g3 = []
        for solid in self.solids:
            g3.append(WallBoundary(solid, sources=self.fluids))
        equations.append(Group(equations=g3))

        if self.has_ghosts:
            gh = []
            for fluid in self.fluids:
                gh.append(
                    MPMUpdateGhostProps(dest=fluid, sources=None)
                )
            equations.append(Group(equations=gh, real=False))

        g4 = []
        for fluid in self.fluids:
            g4.append(MPMAccelerations(
                dest=fluid, sources=self.fluids + self.solids,
                alpha1_min=self.alpha1,
                alpha2_min=self.alpha2, beta=self.beta,
                update_alpha1=self.update_alpha1,
                update_alpha2=self.update_alpha2
            ))
        equations.append(Group(equations=g4))

        return equations

    def setup_properties(self, particles, clean=True):
        from pysph.base.utils import get_particle_array_gasd
        import numpy
        particle_arrays = dict([(p.name, p) for p in particles])
        dummy = get_particle_array_gasd(name='junk')
        props = list(dummy.properties.keys())
        output_props = dummy.output_property_arrays
        for fluid in self.fluids:
            pa = particle_arrays[fluid]
            self._ensure_properties(pa, props, clean)
            pa.add_property('orig_idx', type='int')
            nfp = pa.get_number_of_particles()
            pa.orig_idx[:] = numpy.arange(nfp)
            pa.set_output_arrays(output_props)

        solid_props = set(props) | set('div cs wij htmp'.split(' '))
        for solid in self.solids:
            pa = particle_arrays[solid]
            self._ensure_properties(pa, solid_props, clean)
            pa.set_output_arrays(output_props)


class GSPHScheme(Scheme):
    def __init__(self, fluids, solids, dim, gamma, kernel_factor, g1=0.0,
                 g2=0.0, rsolver=2, interpolation=1, monotonicity=1,
                 interface_zero=True, hybrid=False, blend_alpha=5.0, tf=1.0,
                 niter=20, tol=1e-6, has_ghosts=False):
        """
        Parameters
        ----------

        fluids: list
            List of names of fluid particle arrays.
        solids: list
            List of names of solid particle arrays (or boundaries), currently
            not supported
        dim: int
            Dimensionality of the problem.
        gamma: float
            Gamma for Equation of state.
        kernel_factor: float
            Kernel scaling factor.
        g1, g2 : double
            ADKE style thermal conduction parameters
        rsolver: int
            Riemann solver to use.  See pysph.sph.gas_dynamics.gsph for
            valid options.
        interpolation: int
            Kind of interpolation for the specific volume integrals.
        monotonicity : int
            Type of monotonicity algorithm to use:
            0 : First order GSPH
            1 : I02 algorithm
            2 : IwIn algorithm
        interface_zero : bool
            Set Interface position s^*_{ij} = 0 for the Riemann problem.
        hybrid, blend_alpha : bool, double
            Hybrid scheme and blending alpha value
        tf: double
            Final time used for blending.
        niter: int
            Max number of iterations for iterative Riemann solvers.
        tol: double
            Tolerance for iterative Riemann solvers.
        has_ghosts: bool
            if ghost particles (either mirror or periodic) is used
        """
        self.fluids = fluids
        self.solids = solids
        self.dim = dim
        self.solver = None
        self.gamma = gamma
        self.kernel_factor = kernel_factor
        self.g1 = g1
        self.g2 = g2
        self.rsolver = rsolver
        self.interpolation = interpolation
        self.monotonicity = monotonicity
        self.interface_zero = interface_zero
        self.hybrid = hybrid
        self.blend_alpha = blend_alpha
        self.tf = tf
        self.niter = niter
        self.tol = tol
        self.has_ghosts = has_ghosts

    def add_user_options(self, group):
        group.add_argument(
            "--rsolver", action="store", type=int, dest="rsolver",
            default=None,
            help="Riemann solver to use."
        )
        group.add_argument(
            "--interpolation", action="store", type=int, dest="interpolation",
            default=None,
            help="Interpolation algorithm to use."
        )
        group.add_argument(
            "--monotonicity", action="store", type=int, dest="monotonicity",
            default=None,
            help="Monotonicity algorithm to use."
        )
        group.add_argument(
            "--g1", action="store", type=float, dest="g1",
            default=None,
            help="ADKE style thermal conduction parameter."
        )
        group.add_argument(
            "--g2", action="store", type=float, dest="g2",
            default=None,
            help="ADKE style thermal conduction parameter."
        )
        group.add_argument(
            "--gamma", action="store", type=float, dest="gamma",
            default=None,
            help="Gamma for the state equation."
        )
        group.add_argument(
            "--blend-alpha", action="store", type=float, dest="blend_alpha",
            default=None,
            help="Blending factor for hybrid scheme."
        )
        add_bool_argument(
            group, "interface-zero", dest="interface_zero",
            help="Set interface position to zero for Riemann problem.",
            default=None
        )
        add_bool_argument(
            group, "hybrid", dest="hybrid",
            help="Use the hybrid scheme.",
            default=None
        )

    def consume_user_options(self, options):
        vars = ['gamma', 'g1', 'g2', 'rsolver', 'interpolation',
                'monotonicity', 'interface_zero', 'hybrid',
                'blend_alpha']
        data = dict((var, self._smart_getattr(options, var))
                    for var in vars)
        self.configure(**data)

    def configure_solver(self, kernel=None, integrator_cls=None,
                         extra_steppers=None, **kw):
        """Configure the solver to be generated.

        Parameters
        ----------

        kernel : Kernel instance.
            Kernel to use, if none is passed a default one is used.
        integrator_cls : pysph.sph.integrator.Integrator
            Integrator class to use, use sensible default if none is
            passed.
        extra_steppers : dict
            Additional integration stepper instances as a dict.
        **kw : extra arguments
            Any additional keyword args are passed to the solver instance.
        """
        from pysph.base.kernels import Gaussian
        if kernel is None:
            kernel = Gaussian(dim=self.dim)

        steppers = {}
        if extra_steppers is not None:
            steppers.update(extra_steppers)

        from pysph.sph.integrator import EulerIntegrator
        from pysph.sph.integrator_step import GSPHStep

        cls = integrator_cls if integrator_cls is not None else EulerIntegrator
        step_cls = GSPHStep
        for name in self.fluids:
            if name not in steppers:
                steppers[name] = step_cls()

        integrator = cls(**steppers)

        from pysph.solver.solver import Solver
        self.solver = Solver(
            dim=self.dim, integrator=integrator, kernel=kernel, **kw
        )
        if 'tf' in kw:
            self.tf = kw['tf']

    def get_equations(self):
        from pysph.sph.equation import Group
        from pysph.sph.gas_dynamics.basic import (
            ScaleSmoothingLength, UpdateSmoothingLengthFromVolume,
            SummationDensity, IdealGasEOS
        )
        from pysph.sph.gas_dynamics.boundary_equations import WallBoundary
        from pysph.sph.gas_dynamics.gsph import (
            GSPHGradients, GSPHAcceleration, GSPHUpdateGhostProps
        )
        equations = []
        # Find the optimal 'h'
        group = []
        for fluid in self.fluids:
            group.append(
                ScaleSmoothingLength(dest=fluid, sources=None, factor=2.0)
            )
        equations.append(Group(equations=group, update_nnps=True))

        if self.solids:
            group = []
            for solid in self.solids:
                group.append(WallBoundary(solid, sources=self.fluids))
            equations.append(Group(equations=group))

        all_pa = self.fluids + self.solids
        group = []
        for fluid in self.fluids:
            group.append(
                SummationDensity(
                    dest=fluid, sources=all_pa, dim=self.dim
                )
            )
        equations.append(Group(equations=group, update_nnps=False))

        if self.solids:
            group = []
            for solid in self.solids:
                group.append(WallBoundary(solid, sources=self.fluids))
            equations.append(Group(equations=group))

        group = []
        for fluid in self.fluids:
            group.append(
                UpdateSmoothingLengthFromVolume(
                    dest=fluid, sources=None, k=self.kernel_factor,
                    dim=self.dim
                )
            )
        equations.append(Group(equations=group, update_nnps=True))

        group = []
        for fluid in self.fluids:
            group.append(
                SummationDensity(
                    dest=fluid, sources=all_pa, dim=self.dim
                )
            )
        equations.append(Group(equations=group, update_nnps=False))
        # Done with finding the optimal 'h'

        group = []
        for fluid in self.fluids:
            group.append(IdealGasEOS(dest=fluid, sources=None,
                                     gamma=self.gamma))
        equations.append(Group(equations=group))

        if self.solids:
            group = []
            for solid in self.solids:
                group.append(WallBoundary(solid, sources=self.fluids))
            equations.append(Group(equations=group))

        g2 = []
        for fluid in self.fluids:
            g2.append(GSPHGradients(dest=fluid, sources=all_pa))

        equations.append(Group(equations=g2))

        if self.has_ghosts:
            g3 = []
            for fluid in self.fluids:
                g3.append(GSPHUpdateGhostProps(dest=fluid, sources=None))
            equations.append(Group(
                equations=g3, update_nnps=False, real=False
                ))

        g4 = []
        for fluid in self.fluids:
            g4.append(GSPHAcceleration(
                dest=fluid, sources=all_pa, g1=self.g1,
                g2=self.g2, monotonicity=self.monotonicity,
                rsolver=self.rsolver, interpolation=self.interpolation,
                interface_zero=self.interface_zero,
                hybrid=self.hybrid, blend_alpha=self.blend_alpha,
                gamma=self.gamma, niter=self.niter, tol=self.tol
            ))
        equations.append(Group(equations=g4))
        return equations

    def setup_properties(self, particles, clean=True):
        from pysph.base.utils import get_particle_array_gasd
        import numpy
        particle_arrays = dict([(p.name, p) for p in particles])
        dummy = get_particle_array_gasd(name='junk')
        props = (list(dummy.properties.keys()) +
                 'px py pz ux uy uz vx vy vz wx wy wz'.split())

        output_props = dummy.output_property_arrays
        for fluid in self.fluids:
            pa = particle_arrays[fluid]
            self._ensure_properties(pa, props, clean)
            pa.add_property('orig_idx', type='int')
            nfp = pa.get_number_of_particles()
            pa.orig_idx[:] = numpy.arange(nfp)
            pa.set_output_arrays(output_props)

        solid_props = set(props) | set(('wij', 'htmp'))
        for solid in self.solids:
            pa = particle_arrays[solid]
            self._ensure_properties(pa, solid_props, clean)
            pa.set_output_arrays(output_props)


class ADKEScheme(Scheme):
    def __init__(self, fluids, solids, dim, gamma=1.4, alpha=1.0, beta=2.0,
                 k=1.0, eps=0.0, g1=0, g2=0, has_ghosts=False):
        """
        Parameters
        ----------

        fluids: list
            a list with names of fluid particle arrays
        solids: list
            a list with names of solid (or boundary) particle arrays
        dim: int
            dimensionality of the problem
        gamma: double
            Gamma for equation of state
        alpha: double
            artificial viscosity parameter
        beta: double
            artificial viscosity parameter
        k: double
            kernel scaling parameter
        eps: double
            kernel scaling parameter
        g1: double
            artificial heat conduction parameter
        g2: double
            artificial heat conduction parameter
        has_ghosts: bool
            if problem uses ghost particles (periodic or mirror)
        """
        self.fluids = fluids
        self.solids = solids
        self.dim = dim
        self.solver = None
        self.gamma = gamma
        self.alpha = alpha
        self.beta = beta
        self.k = k
        self.eps = eps
        self.g1 = g1
        self.g2 = g2
        self.has_ghosts = has_ghosts

    def get_equations(self):
        from pysph.sph.equation import Group
        from pysph.sph.basic_equations import SummationDensity
        from pysph.sph.gas_dynamics.basic import (
            IdealGasEOS, ADKEAccelerations, SummationDensityADKE,
            ADKEUpdateGhostProps
        )
        from pysph.sph.gas_dynamics.boundary_equations import WallBoundary

        equations = []
        if self.solids:
            g1 = []
            for solid in self.solids:
                g1.append(WallBoundary(solid, sources=self.fluids))
            equations.append(Group(equations=g1))

        g2 = []
        for fluid in self.fluids:
            g2.append(
                SummationDensityADKE(
                    fluid, sources=self.fluids + self.solids, k=self.k,
                    eps=self.eps
                )
            )
        equations.append(Group(g2, update_nnps=False, iterate=False))

        if self.solids:
            g3 = []
            for solid in self.solids:
                g3.append(WallBoundary(solid, sources=self.fluids))
            equations.append(Group(equations=g3))

        g4 = []
        for fluid in self.fluids:
            g4.append(SummationDensity(fluid, self.fluids+self.solids))
        equations.append(Group(g4,  update_nnps=True))

        if self.solids:
            g5 = []
            for solid in self.solids:
                g5.append(WallBoundary(solid, sources=self.fluids))
            equations.append(Group(equations=g5))

        g6 = []
        for elem in self.fluids+self.solids:
            g6.append(IdealGasEOS(elem, sources=None, gamma=self.gamma))
        equations.append(Group(equations=g6))

        if self.has_ghosts:
            gh = []
            for fluid in self.fluids:
                gh.append(
                    ADKEUpdateGhostProps(dest=fluid, sources=None)
                )
            equations.append(Group(equations=gh, real=False))

        g7 = []
        for fluid in self.fluids:
            g7.append(
                ADKEAccelerations(
                    dest=fluid, sources=self.fluids + self.solids,
                    alpha=self.alpha, beta=self.beta, g1=self.g1, g2=self.g2,
                    k=self.k, eps=self.eps
                )
            )

        equations.append(Group(equations=g7))
        return equations

    def configure_solver(self, kernel=None, integrator_cls=None,
                         extra_steppers=None, **kw):
        """Configure the solver to be generated.

        Parameters
        ----------

        kernel : Kernel instance.
            Kernel to use, if none is passed a default one is used.
        integrator_cls : pysph.sph.integrator.Integrator
            Integrator class to use, use sensible default if none is
            passed.
        extra_steppers : dict
            Additional integration stepper instances as a dict.
        **kw : extra arguments
            Any additional keyword args are passed to the solver instance.
        """
        from pysph.base.kernels import Gaussian
        if kernel is None:
            kernel = Gaussian(dim=self.dim)

        steppers = {}
        if extra_steppers is not None:
            steppers.update(extra_steppers)

        from pysph.sph.integrator import PECIntegrator
        from pysph.sph.integrator_step import ADKEStep

        cls = integrator_cls if integrator_cls is not None else PECIntegrator
        step_cls = ADKEStep
        for name in self.fluids:
            if name not in steppers:
                steppers[name] = step_cls()

        integrator = cls(**steppers)

        from pysph.solver.solver import Solver
        self.solver = Solver(
            dim=self.dim, integrator=integrator, kernel=kernel, **kw
        )

    def setup_properties(self, particles, clean=True):
        from pysph.base.utils import get_particle_array
        import numpy
        particle_arrays = dict([(p.name, p) for p in particles])
        required_props = [
                'x', 'y', 'z', 'u', 'v', 'w', 'rho', 'h', 'm', 'cs', 'p',
                'e', 'au', 'av', 'aw', 'arho', 'ae', 'am', 'ah', 'x0', 'y0',
                'z0', 'u0', 'v0', 'w0', 'rho0', 'e0', 'h0', 'div',  'h0',
                'wij', 'htmp', 'logrho']

        dummy = get_particle_array(additional_props=required_props,
                                   name='junk')
        dummy.set_output_arrays(
            ['x', 'y', 'u', 'v', 'rho', 'm', 'h',
             'cs', 'p', 'e', 'au', 'av', 'ae', 'pid', 'gid', 'tag']
        )

        props = list(dummy.properties.keys())
        output_props = dummy.output_property_arrays
        for solid in self.solids:
            pa = particle_arrays[solid]
            self._ensure_properties(pa, props, clean)
            pa.set_output_arrays(output_props)

        for fluid in self.fluids:
            pa = particle_arrays[fluid]
            self._ensure_properties(pa, props, clean)
            pa.add_property('orig_idx', type='int')
            nfp = pa.get_number_of_particles()
            pa.orig_idx[:] = numpy.arange(nfp)
            pa.set_output_arrays(output_props)



# cython: language_level=3, embedsignature=True
# distutils: language=c++
from libcpp.vector cimport vector
from libcpp.map cimport map
from libcpp.pair cimport pair

from pysph.base.gpu_nnps_base cimport *

ctypedef unsigned int u_int
ctypedef map[u_int, pair[u_int, u_int]] key_to_idx_t
ctypedef vector[u_int] u_int_vector_t

cdef extern from 'math.h':
    int abs(int) nogil
    double ceil(double) nogil
    double floor(double) nogil
    double fabs(double) nogil
    double fmax(double, double) nogil
    double fmin(double, double) nogil

cdef extern from 'math.h':
    double log(double) nogil
    double log2(double) nogil

cdef class StratifiedSFCGPUNNPS(GPUNNPS):
    cdef NNPSParticleArrayWrapper src, dst # Current source and destination.

    cdef public list pids
    cdef public list pid_keys
    cdef public list start_idx_levels
    cdef public list num_particles_levels
    cdef public int max_num_bits
    cdef int num_levels
    cdef double interval_size
    cdef double eps

    cdef object helper

    cdef bint _sorted

    cpdef get_spatially_ordered_indices(self, int pa_index)

    cpdef _bin(self, int pa_index)

    cpdef _refresh(self)

    cdef void find_neighbor_lengths(self, nbr_lengths)

    cdef void find_nearest_neighbors_gpu(self, nbrs, start_indices)



import numpy as np
try:
    from scipy.integrate import quad
except ImportError:
    quad = None
from unittest import TestCase, main

from pysph.base.kernels import (CubicSpline, Gaussian, QuinticSpline,
                                SuperGaussian, WendlandQuintic,
                                WendlandQuinticC4, WendlandQuinticC6,
                                WendlandQuinticC2_1D, WendlandQuinticC4_1D,
                                WendlandQuinticC6_1D, get_compiled_kernel)


###############################################################################
# `TestKernelBase` class.
###############################################################################
class TestKernelBase(TestCase):
    """Base class for all kernel tests.
    """
    kernel_factory = None

    @classmethod
    def setUpClass(cls):
        cls.wrapper = get_compiled_kernel(cls.kernel_factory())
        cls.kernel = cls.wrapper.kernel
        cls.gradient = cls.wrapper.gradient

    def setUp(self):
        self.kernel = self.__class__.kernel
        self.gradient = self.__class__.gradient

    def check_kernel_moment_1d(self, a, b, h, m, xj=0.0):
        func = self.kernel
        if m == 0:
            def f(x): return func(x, 0, 0, xj, 0, 0, h)
        else:
            def f(x): return (pow(x, m) * func(x, 0, 0, xj, 0, 0, h))
        if quad is None:
            kern_f = np.vectorize(f)
            nx = 201
            x = np.linspace(a, b, nx)
            result = np.sum(kern_f(x)) * (b - a) / (nx - 1)
        else:
            result = quad(f, a, b)[0]

        return result

    def check_grad_moment_1d(self, a, b, h, m, xj=0.0):
        func = self.gradient
        if m == 0:
            def f(x): return func(x, 0, 0, xj, 0, 0, h)[0]
        else:
            def f(x): return (pow(x - xj, m) * func(x, 0, 0, xj, 0, 0, h)[0])
        if quad is None:
            kern_f = np.vectorize(f)
            nx = 201
            x = np.linspace(a, b, nx)
            return np.sum(kern_f(x)) * (b - a) / (nx - 1)
        else:
            return quad(f, a, b)[0]

    def check_kernel_moment_2d(self, m, n):
        x0, y0, z0 = 0.5, 0.5, 0.0

        def func(x, y):
            fac = pow(x - x0, m) * pow(y - y0, n)
            return fac * self.kernel(x, y, 0.0, x0, y0, 0.0, 0.15)
        vfunc = np.vectorize(func)
        nx, ny = 101, 101
        vol = 1.0 / (nx - 1) * 1.0 / (ny - 1)
        x, y = np.mgrid[0:1:nx * 1j, 0:1:nx * 1j]
        result = np.sum(vfunc(x, y)) * vol
        return result

    def check_gradient_moment_2d(self, m, n):
        x0, y0, z0 = 0.5, 0.5, 0.0

        def func(x, y):
            fac = pow(x - x0, m) * pow(y - y0, n)
            return fac*np.asarray(self.gradient(x, y, 0.0, x0, y0, 0.0, 0.15))
        vfunc = np.vectorize(func, otypes=[np.ndarray])
        nx, ny = 101, 101
        vol = 1.0 / (nx - 1) * 1.0 / (ny - 1)
        x, y = np.mgrid[0:1:nx * 1j, 0:1:nx * 1j]
        result = np.sum(vfunc(x, y)) * vol
        return result

    def check_kernel_moment_3d(self, l, m, n):
        x0, y0, z0 = 0.5, 0.5, 0.5

        def func(x, y, z):
            fac = pow(x - x0, l) * pow(y - y0, m) * pow(z - z0, n)
            return fac * self.kernel(x, y, z, x0, y0, z0, 0.15)
        vfunc = np.vectorize(func)
        nx, ny, nz = 51, 51, 51
        vol = 1.0 / (nx - 1) * 1.0 / (ny - 1) * 1.0 / (nz - 1)
        x, y, z = np.mgrid[0:1:nx * 1j, 0:1:ny * 1j, 0:1:nz * 1j]
        result = np.sum(vfunc(x, y, z)) * vol
        return result

    def check_gradient_moment_3d(self, l, m, n):
        x0, y0, z0 = 0.5, 0.5, 0.5

        def func(x, y, z):
            fac = pow(x - x0, l) * pow(y - y0, m) * pow(z - z0, n)
            return fac * np.asarray(self.gradient(x, y, z, x0, y0, z0, 0.15))
        vfunc = np.vectorize(func, otypes=[np.ndarray])
        nx, ny, nz = 51, 51, 51
        vol = 1.0 / (nx - 1) * 1.0 / (ny - 1) * 1.0 / (nz - 1)
        x, y, z = np.mgrid[0:1:nx * 1j, 0:1:ny * 1j, 0:1:nz * 1j]
        result = np.sum(vfunc(x, y, z)) * vol
        return result

    def check_kernel_at_origin(self, w_0):
        k = self.kernel(xi=0.0, yi=0.0, zi=0.0, xj=0.0, yj=0.0, zj=0.0, h=1.0)
        expect = w_0
        self.assertAlmostEqual(k, expect,
                               msg='Kernel value %s != %s (expected)'
                               % (k, expect))
        k = self.kernel(xi=3.0, yi=0.0, zi=0.0, xj=0.0, yj=0.0, zj=0.0, h=1.0)
        expect = 0.0
        self.assertAlmostEqual(k, expect,
                               msg='Kernel value %s != %s (expected)'
                               % (k, expect))

        g = self.gradient(xi=0.0, yi=0.0, zi=0.0,
                          xj=0.0, yj=0.0, zj=0.0, h=1.0)
        expect = 0.0
        self.assertAlmostEqual(g[0], expect,
                               msg='Kernel value %s != %s (expected)'
                               % (g[0], expect))
        g = self.gradient(xi=3.0, yi=0.0, zi=0.0,
                          xj=0.0, yj=0.0, zj=0.0, h=1.0)
        expect = 0.0
        self.assertAlmostEqual(g[0], expect,
                               msg='Kernel value %s != %s (expected)'
                               % (g[0], expect))


###############################################################################
# `TestCubicSpline1D` class.
###############################################################################
class TestCubicSpline1D(TestKernelBase):
    kernel_factory = staticmethod(lambda: CubicSpline(dim=1))

    def test_simple(self):
        self.check_kernel_at_origin(2. / 3)

    def test_zeroth_kernel_moments(self):
        kh = self.wrapper.radius_scale
        # zero'th moment
        r = self.check_kernel_moment_1d(-kh, kh, 1.0, 0, xj=0)
        self.assertAlmostEqual(r, 1.0, 8)
        # Use a non-unit h.
        r = self.check_kernel_moment_1d(-kh, kh, 0.5, 0, xj=0)
        self.assertAlmostEqual(r, 1.0, 8)
        r = self.check_kernel_moment_1d(0.0, 2 * kh, 1.0, 0, xj=kh)
        self.assertAlmostEqual(r, 1.0, 8)

    def test_first_kernel_moment(self):
        kh = self.wrapper.radius_scale
        r = self.check_kernel_moment_1d(-kh, kh, 1.0, 1, xj=0.0)
        self.assertAlmostEqual(r, 0.0, 8)

    def test_zeroth_grad_moments(self):
        kh = self.wrapper.radius_scale
        # zero'th moment
        r = self.check_grad_moment_1d(-kh, kh, 1.0, 0, xj=0)
        self.assertAlmostEqual(r, 0.0, 8)
        # Use a non-unit h.
        r = self.check_grad_moment_1d(-kh, kh, 0.5, 0, xj=0)
        self.assertAlmostEqual(r, 0.0, 8)
        r = self.check_grad_moment_1d(0.0, 2 * kh, 1.0, 0, xj=kh)
        self.assertAlmostEqual(r, 0.0, 8)

    def test_first_grad_moment(self):
        kh = self.wrapper.radius_scale
        r = self.check_grad_moment_1d(0.0, 2 * kh, 1.0, 1, xj=kh)
        self.assertAlmostEqual(r, -1.0, 8)

###############################################################################
# `TestCubicSpline2D` class.
###############################################################################


class TestCubicSpline2D(TestKernelBase):
    kernel_factory = staticmethod(lambda: CubicSpline(dim=2))

    def test_simple(self):
        self.check_kernel_at_origin(10. / (7 * np.pi))

    def test_zeroth_kernel_moments(self):
        r = self.check_kernel_moment_2d(0, 0)
        self.assertAlmostEqual(r, 1.0, 7)

    def test_first_kernel_moment(self):
        r = self.check_kernel_moment_2d(0, 1)
        self.assertAlmostEqual(r, 0.0, 7)
        r = self.check_kernel_moment_2d(1, 0)
        self.assertAlmostEqual(r, 0.0, 7)
        r = self.check_kernel_moment_2d(1, 1)
        self.assertAlmostEqual(r, 0.0, 7)

    def test_zeroth_grad_moments(self):
        r = self.check_gradient_moment_2d(0, 0)
        self.assertAlmostEqual(r[0], 0.0, 7)
        self.assertAlmostEqual(r[1], 0.0, 7)

    def test_first_grad_moment(self):
        r = self.check_gradient_moment_2d(1, 0)
        self.assertAlmostEqual(r[0], -1.0, 6)
        self.assertAlmostEqual(r[1], 0.0, 8)
        r = self.check_gradient_moment_2d(0, 1)
        self.assertAlmostEqual(r[0], 0.0, 8)
        self.assertAlmostEqual(r[1], -1.0, 6)
        r = self.check_gradient_moment_2d(1, 1)
        self.assertAlmostEqual(r[0], 0.0, 8)
        self.assertAlmostEqual(r[1], 0.0, 8)


###############################################################################
# `TestCubicSpline3D` class.
###############################################################################
class TestCubicSpline3D(TestKernelBase):
    kernel_factory = staticmethod(lambda: CubicSpline(dim=3))

    def test_simple(self):
        self.check_kernel_at_origin(1. / np.pi)

    def test_zeroth_kernel_moments(self):
        r = self.check_kernel_moment_3d(0, 0, 0)
        self.assertAlmostEqual(r, 1.0, 6)

    def test_first_kernel_moment(self):
        r = self.check_kernel_moment_3d(0, 0, 1)
        self.assertAlmostEqual(r, 0.0, 7)
        r = self.check_kernel_moment_3d(0, 1, 0)
        self.assertAlmostEqual(r, 0.0, 7)
        r = self.check_kernel_moment_3d(1, 0, 1)
        self.assertAlmostEqual(r, 0.0, 7)

    def test_zeroth_grad_moments(self):
        r = self.check_gradient_moment_3d(0, 0, 0)
        self.assertAlmostEqual(r[0], 0.0, 7)
        self.assertAlmostEqual(r[1], 0.0, 7)
        self.assertAlmostEqual(r[2], 0.0, 7)

    def test_first_grad_moment(self):
        r = self.check_gradient_moment_3d(1, 0, 0)
        self.assertAlmostEqual(r[0], -1.0, 4)
        self.assertAlmostEqual(r[1], 0.0, 8)
        self.assertAlmostEqual(r[2], 0.0, 8)
        r = self.check_gradient_moment_3d(0, 1, 0)
        self.assertAlmostEqual(r[0], 0.0, 8)
        self.assertAlmostEqual(r[1], -1.0, 4)
        self.assertAlmostEqual(r[2], 0.0, 6)
        r = self.check_gradient_moment_3d(0, 0, 1)
        self.assertAlmostEqual(r[0], 0.0, 8)
        self.assertAlmostEqual(r[1], 0.0, 8)
        self.assertAlmostEqual(r[2], -1.0, 4)

###############################################################################
# Gaussian kernel


class TestGaussian1D(TestCubicSpline1D):
    kernel_factory = staticmethod(lambda: Gaussian(dim=1))

    def test_simple(self):
        self.check_kernel_at_origin(1.0 / np.sqrt(np.pi))

    def test_first_grad_moment(self):
        kh = self.wrapper.radius_scale
        r = self.check_grad_moment_1d(0.0, 2 * kh, 1.0, 1, xj=kh)
        self.assertAlmostEqual(r, -1.0, 3)

    def test_zeroth_kernel_moments(self):
        kh = self.wrapper.radius_scale
        # zero'th moment
        r = self.check_kernel_moment_1d(-kh, kh, 1.0, 0, xj=0)
        self.assertAlmostEqual(r, 1.0, 4)
        # Use a non-unit h.
        r = self.check_kernel_moment_1d(-kh, kh, 0.5, 0, xj=0)
        self.assertAlmostEqual(r, 1.0, 4)
        r = self.check_kernel_moment_1d(0.0, 2 * kh, 1.0, 0, xj=kh)
        self.assertAlmostEqual(r, 1.0, 4)


class TestGaussian2D(TestCubicSpline2D):
    kernel_factory = staticmethod(lambda: Gaussian(dim=2))

    def test_simple(self):
        self.check_kernel_at_origin(1.0 / np.pi)

    def test_zeroth_kernel_moments(self):
        r = self.check_kernel_moment_2d(0, 0)
        self.assertAlmostEqual(r, 1.0, 3)

    def test_first_grad_moment(self):
        r = self.check_gradient_moment_2d(1, 0)
        self.assertAlmostEqual(r[0], -1.0, 2)
        self.assertAlmostEqual(r[1], 0.0, 8)
        r = self.check_gradient_moment_2d(0, 1)
        self.assertAlmostEqual(r[0], 0.0, 8)
        self.assertAlmostEqual(r[1], -1.0, 2)
        r = self.check_gradient_moment_2d(1, 1)
        self.assertAlmostEqual(r[0], 0.0, 8)
        self.assertAlmostEqual(r[1], 0.0, 8)


class TestGaussian3D(TestCubicSpline3D):
    kernel_factory = staticmethod(lambda: Gaussian(dim=3))

    def test_simple(self):
        self.check_kernel_at_origin(1.0 / pow(np.pi, 1.5))

    def test_zeroth_kernel_moments(self):
        r = self.check_kernel_moment_3d(0, 0, 0)
        self.assertAlmostEqual(r, 1.0, 2)

    def test_first_grad_moment(self):
        r = self.check_gradient_moment_3d(1, 0, 0)
        self.assertAlmostEqual(r[0], -1.0, 2)
        self.assertAlmostEqual(r[1], 0.0, 8)
        self.assertAlmostEqual(r[2], 0.0, 8)
        r = self.check_gradient_moment_3d(0, 1, 0)
        self.assertAlmostEqual(r[0], 0.0, 8)
        self.assertAlmostEqual(r[1], -1.0, 2)
        self.assertAlmostEqual(r[2], 0.0, 6)
        r = self.check_gradient_moment_3d(0, 0, 1)
        self.assertAlmostEqual(r[0], 0.0, 8)
        self.assertAlmostEqual(r[1], 0.0, 8)
        self.assertAlmostEqual(r[2], -1.0, 2)


###############################################################################
# Quintic spline kernel
class TestQuinticSpline1D(TestCubicSpline1D):
    kernel_factory = staticmethod(lambda: QuinticSpline(dim=1))

    def test_simple(self):
        self.check_kernel_at_origin(0.55)


class TestQuinticSpline2D(TestCubicSpline2D):
    kernel_factory = staticmethod(lambda: QuinticSpline(dim=2))

    def test_simple(self):
        self.check_kernel_at_origin(66.0 * 7.0 / (478.0 * np.pi))


class TestQuinticSpline3D(TestGaussian3D):
    kernel_factory = staticmethod(lambda: QuinticSpline(dim=3))

    def test_simple(self):
        self.check_kernel_at_origin(66.0 * 1.0 / (120.0 * np.pi))


###############################################################################
# SuperGaussian kernel
class TestSuperGaussian1D(TestGaussian1D):
    kernel_factory = staticmethod(lambda: SuperGaussian(dim=1))

    def test_simple(self):
        self.check_kernel_at_origin(1.5 / np.sqrt(np.pi))

    def test_first_grad_moment(self):
        kh = self.wrapper.radius_scale
        r = self.check_grad_moment_1d(0.0, 2 * kh, 1.0, 1, xj=kh)
        self.assertAlmostEqual(r, -1.0, 2)

    def test_zeroth_kernel_moments(self):
        kh = self.wrapper.radius_scale
        # zero'th moment
        r = self.check_kernel_moment_1d(-kh, kh, 1.0, 0, xj=0)
        self.assertAlmostEqual(r, 1.0, 3)
        # Use a non-unit h.
        r = self.check_kernel_moment_1d(-kh, kh, 0.5, 0, xj=0)
        self.assertAlmostEqual(r, 1.0, 3)
        r = self.check_kernel_moment_1d(0.0, 2 * kh, 1.0, 0, xj=kh)
        self.assertAlmostEqual(r, 1.0, 3)


class TestSuperGaussian2D(TestGaussian2D):
    kernel_factory = staticmethod(lambda: SuperGaussian(dim=2))

    def test_simple(self):
        self.check_kernel_at_origin(2.0 / np.pi)

    def test_zeroth_kernel_moments(self):
        r = self.check_kernel_moment_2d(0, 0)
        self.assertAlmostEqual(r, 1.0, 2)

    def test_first_grad_moment(self):
        r = self.check_gradient_moment_2d(1, 0)
        self.assertAlmostEqual(r[0], -1.0, 1)
        self.assertAlmostEqual(r[1], 0.0, 8)
        r = self.check_gradient_moment_2d(0, 1)
        self.assertAlmostEqual(r[0], 0.0, 8)
        self.assertAlmostEqual(r[1], -1.0, 1)
        r = self.check_gradient_moment_2d(1, 1)
        self.assertAlmostEqual(r[0], 0.0, 8)
        self.assertAlmostEqual(r[1], 0.0, 8)


class TestSuperGaussian3D(TestGaussian3D):
    kernel_factory = staticmethod(lambda: SuperGaussian(dim=3))

    def test_simple(self):
        self.check_kernel_at_origin(5.0 / (2.0 * pow(np.pi, 1.5)))

    def test_first_grad_moment(self):
        r = self.check_gradient_moment_3d(1, 0, 0)
        self.assertAlmostEqual(r[0], -1.0, 1)
        self.assertAlmostEqual(r[1], 0.0, 8)
        self.assertAlmostEqual(r[2], 0.0, 8)
        r = self.check_gradient_moment_3d(0, 1, 0)
        self.assertAlmostEqual(r[0], 0.0, 8)
        self.assertAlmostEqual(r[1], -1.0, 1)
        self.assertAlmostEqual(r[2], 0.0, 6)
        r = self.check_gradient_moment_3d(0, 0, 1)
        self.assertAlmostEqual(r[0], 0.0, 8)
        self.assertAlmostEqual(r[1], 0.0, 8)
        self.assertAlmostEqual(r[2], -1.0, 1)

###############################################################################
# WendlandQuintic C2 kernel


class TestWendlandQuintic2D(TestCubicSpline2D):
    kernel_factory = staticmethod(lambda: WendlandQuintic(dim=2))

    def test_simple(self):
        self.check_kernel_at_origin(7.0 / (4.0 * np.pi))

    def test_zeroth_kernel_moments(self):
        r = self.check_kernel_moment_2d(0, 0)
        self.assertAlmostEqual(r, 1.0, 6)


class TestWendlandQuintic3D(TestGaussian3D):
    kernel_factory = staticmethod(lambda: WendlandQuintic(dim=3))

    def test_simple(self):
        self.check_kernel_at_origin(21.0 / (16.0 * np.pi))


class TestWendlandQuintic1D(TestCubicSpline1D):
    kernel_factory = staticmethod(lambda: WendlandQuinticC2_1D(dim=1))

    def test_simple(self):
        self.check_kernel_at_origin(5.0 / 8.0)

    def test_zeroth_kernel_moments(self):
        kh = self.wrapper.radius_scale
        # zero'th moment
        r = self.check_kernel_moment_1d(-kh, kh, 1.0, 0, xj=0)
        self.assertAlmostEqual(r, 1.0, 8)
        # Use a non-unit h.
        r = self.check_kernel_moment_1d(-kh, kh, 0.5, 0, xj=0)
        self.assertAlmostEqual(r, 1.0, 7)
        r = self.check_kernel_moment_1d(0.0, 2 * kh, 1.0, 0, xj=kh)
        self.assertAlmostEqual(r, 1.0, 7)

    def test_first_grad_moment(self):
        kh = self.wrapper.radius_scale
        r = self.check_grad_moment_1d(0.0, 2 * kh, 1.0, 1, xj=kh)
        self.assertAlmostEqual(r, -1.0, 7)


###############################################################################
# WendlandQuintic C4 kernel
class TestWendlandQuinticC4_2D(TestCubicSpline2D):
    kernel_factory = staticmethod(lambda: WendlandQuinticC4(dim=2))

    def test_simple(self):
        self.check_kernel_at_origin(9.0 / (4.0 * np.pi))


class TestWendlandQuinticC4_3D(TestGaussian3D):
    kernel_factory = staticmethod(lambda: WendlandQuinticC4(dim=3))

    def test_simple(self):
        self.check_kernel_at_origin(495.0 / (256.0 * np.pi))


class TestWendlandQuinticC4_1D(TestCubicSpline1D):
    kernel_factory = staticmethod(lambda: WendlandQuinticC4_1D(dim=1))

    def test_simple(self):
        self.check_kernel_at_origin(3.0 / 4.0)


###############################################################################
# WendlandQuintic C6 kernel
class TestWendlandQuinticC6_2D(TestCubicSpline2D):
    kernel_factory = staticmethod(lambda: WendlandQuinticC6(dim=2))

    def test_simple(self):
        self.check_kernel_at_origin(78.0 / (28.0 * np.pi))


class TestWendlandQuinticC6_3D(TestGaussian3D):
    kernel_factory = staticmethod(lambda: WendlandQuinticC6(dim=3))

    def test_simple(self):
        self.check_kernel_at_origin(1365.0 / (512.0 * np.pi))


class TestWendlandQuinticC6_1D(TestCubicSpline1D):
    kernel_factory = staticmethod(lambda: WendlandQuinticC6_1D(dim=1))

    def test_simple(self):
        self.check_kernel_at_origin(55.0 / 64.0)


if __name__ == '__main__':
    main()



"""Make a directory compatible with mybinder.org and ready for
upload to a Github repo.
"""

from pysph.solver.utils import get_files, mkdir
import os
from shutil import copy
import nbformat as nbf
import argparse
import sys
import re
import glob
from textwrap import dedent


def find_viewer_type(path):
    '''
    Finds the type of viewer to use in the jupyter notebook.
    Parses the log file at the file path and searches for
    'dim=d', where 'dD' is taken to be the viewer type.

    example: if 'dim=2', then what is returned is '2D'
    '''

    log_file_path = os.path.abspath(path) + '/*.log'
    regex = r'dim=(\d)'
    log_files = glob.glob(log_file_path)
    if not log_files:
        return '2D'

    match_list = []
    with open(log_files[0], 'r') as file:
        for line in file:
            for match in re.finditer(regex, line, re.S):
                match_text = match.group()
                match_list.append(match_text)
                if len(match_list) > 0:
                    break
            if len(match_list) > 0:
                break
    return match_list[0][-1] + 'D'


def make_notebook(path, sim_name, config_dict={}):
    '''
    Makes a jupyter notebook to view simulation results stored in
    a given directory

    path: the directory conatining the output files

    sim_name: name of the simulation
              ex. 'cavity_output'

    config_dict: configuration dictionary for the notbeook viewer [dict]
                 ex. {'fluid': {'frame': 20}}
    '''

    fname = os.path.join(path, sim_name+'.ipynb')
    if os.path.exists(fname):
        print('Jupyter notebook file {} already exists!'.format(fname))
        print('Remove it to regenerate.')
        return

    viewer_type = find_viewer_type(path)
    cell1_src = [
        "import os\n",
        "from pysph.tools.ipy_viewer import Viewer" + viewer_type
    ]
    cell2_src = [
        "cwd = os.getcwd()"
    ]
    cell3_src = [
        "viewer = Viewer" + viewer_type + "(cwd)\n",
        "viewer.interactive_plot(" + str(config_dict) + ")"
    ]

    nb = nbf.v4.new_notebook()

    nb.cells = [
        nbf.v4.new_code_cell(source=cell1_src),
        nbf.v4.new_code_cell(source=cell2_src),
        nbf.v4.new_code_cell(source=cell3_src)
    ]

    nbf.write(nb, fname)
    return


def find_sim_dirs(path, sim_paths_list=None):
    '''
    Finds all the directories in a given directory that
    contain pysph output files.
    '''
    if sim_paths_list is None:
        sim_paths_list = []
    path = os.path.abspath(path)
    sim_files = get_files(path)
    if len(sim_files) != 0:
        sim_paths_list.append(path)
    elif len(sim_files) == 0:
        files = os.listdir(path)
        files = [f for f in files if os.path.isdir(f)]
        files = [os.path.join(path, f) for f in files if not f.startswith('.')]
        for f in files:
            sim_paths_list = find_sim_dirs(f, sim_paths_list)

    return sim_paths_list


def find_dir_size(path):
    '''
    Finds the size of a given directory.
    '''
    total_size = 0
    for dir_path, dir_names, file_names in os.walk(path):
        for f in file_names:
            fp = os.path.join(dir_path, f)
            if not os.path.islink(fp):
                total_size += os.path.getsize(fp)

    return total_size


def _write_readme_requrements(path):
    reqs = os.path.join(path, 'requirements.txt')
    if os.path.exists(reqs):
        print('Requirements file {} already exists!'.format(reqs))
        print('Remove it to regenerate.')
        return
    else:
        with open(reqs, 'w') as file:
            file.write(
                dedent('''\
                ipympl
                matplotlib
                ipyvolume
                numpy
                pytools
                Beaker
                mako
                Cython
                h5py
                meshio
                cyarray
                compyle
                https://github.com/pypr/pysph/zipball/master
                ''')
            )

    readme = os.path.join(path, 'README.md')
    if os.path.exists(readme):
        print("README exists, not clobbering.")
    else:
        with open(os.path.join(path, 'README.md'), 'w') as file:
            file.write(
                dedent('''\
            # Title
            [![Binder](https://mybinder.org/badge_logo.svg)]
            (https://mybinder.org/v2/gh/user_name/repo_name/branch_name)

            [comment]: # (The above link is for repositories hosted on GitHub.
            For links corresponding to other hosting services,
            please visit https://mybinder.org or
            https://mybinder.readthedocs.io)
            ''')
            )


def make_binder(path):

    src_path = os.path.abspath(path)

    sim_paths_list = find_sim_dirs(src_path)

    for path in sim_paths_list:
        sim_name = os.path.split(path)[1]
        make_notebook(path, sim_name)
        if len(sim_paths_list) == 1 and sim_paths_list[0] == src_path:
            files = os.listdir(src_path)
            files = [os.path.join(src_path, f) for f in files]
            mkdir(os.path.join(src_path, sim_name))
            could_not_copy = []
            for f in files:
                try:
                    copy(f, os.path.join(src_path, sim_name))
                except BaseException as exc:
                    could_not_copy.append([f, exc])
                    continue
                os.remove(f)
            if len(could_not_copy) != 0:
                print("Could not copy the following files:\n")
                for f in could_not_copy:
                    print('file: ', f[0])
                    print('error: ', f[1], '\n')

    _write_readme_requrements(src_path)


def main(argv=None):
    if argv is None:
        argv = sys.argv[1:]

    parser = argparse.ArgumentParser(
        prog='binder',
        description=__doc__,
        add_help=False
    )

    parser.add_argument(
        "-h",
        "--help",
        action="store_true",
        default=False,
        dest="help",
        help="show this help message and exit"
    )

    parser.add_argument(
        "src_path",
        type=str,
        nargs=1,
        help="the directory containing the directories/files to be prepared"
    )

    if len(argv) > 0 and argv[0] in ['-h', '--help']:
        parser.print_help()
        sys.exit()

    options, extra = parser.parse_known_args(argv)

    src_path = options.src_path[0]

    make_binder(src_path)


if __name__ == '__main__':
    main()



"""
Transport Velocity Formulation
##############################

References
----------
    .. [Adami2012] S. Adami et. al "A generalized wall boundary condition for
        smoothed particle hydrodynamics", Journal of Computational Physics
        (2012), pp. 7057--7075.

    .. [Adami2013] S. Adami et. al "A transport-velocity formulation for
        smoothed particle hydrodynamics", Journal of Computational Physics
        (2013), pp. 292--307.

"""

from pysph.sph.equation import Equation
from math import sin, pi

# constants
M_PI = pi


class SummationDensity(Equation):
    r"""**Summation density with volume summation**

    In addition to the standard summation density, the number density
    for the particle is also computed. The number density is important
    for multi-phase flows to define a local particle volume
    independent of the material density.

    .. math::

        \rho_a = \sum_b m_b W_{ab}\\

        \mathcal{V}_a = \frac{1}{\sum_b W_{ab}}

    Notes
    -----

    Note that in the pysph implementation, V is the inverse volume of a
    particle, i.e. the equation computes V as follows:

    .. math::

       \mathcal{V}_a = \sum_b W_{ab}

    For this equation, the destination particle array must define the
    variable `V` for particle volume.
    """

    def initialize(self, d_idx, d_V, d_rho):
        d_V[d_idx] = 0.0
        d_rho[d_idx] = 0.0

    def loop(self, d_idx, d_V, d_rho, d_m, WIJ):
        d_V[d_idx] += WIJ
        d_rho[d_idx] += d_m[d_idx]*WIJ


class VolumeSummation(Equation):
    r"""**Number density for volume computation**

    See `SummationDensity`

    Note that the quantity `V` is really :math:`\sigma` of the original paper,
    i.e. inverse of the particle volume.

    """

    def initialize(self, d_idx, d_V):
        d_V[d_idx] = 0.0

    def loop(self, d_idx, d_V, WIJ):
        d_V[d_idx] += WIJ


class VolumeFromMassDensity(Equation):
    """**Set the inverse volume using mass density**"""
    def loop(self, d_idx, d_V, d_rho, d_m):
        d_V[d_idx] = d_rho[d_idx]/d_m[d_idx]


class SetWallVelocity(Equation):
    r"""**Extrapolating the fluid velocity on to the wall**

    Eq. (22) in [Adami2012]:

    .. math::

        \tilde{\boldsymbol{v}}_a = \frac{\sum_b\boldsymbol{v}_b W_{ab}}
        {\sum_b W_{ab}}

    Notes
    -----
    The destination particle array for this equation should define the
    *filtered* velocity variables :math:`uf, vf, wf`.

    """
    def initialize(self, d_idx, d_uf, d_vf, d_wf, d_wij):
        d_uf[d_idx] = 0.0
        d_vf[d_idx] = 0.0
        d_wf[d_idx] = 0.0
        d_wij[d_idx] = 0.0

    def loop(self, d_idx, s_idx, d_uf, d_vf, d_wf,
             s_u, s_v, s_w, d_wij, WIJ):

        # normalisation factor is different from 'V' as the particles
        # near the boundary do not have full kernel support
        d_wij[d_idx] += WIJ

        # sum in Eq. (22)
        # this will be normalized in post loop
        d_uf[d_idx] += s_u[s_idx] * WIJ
        d_vf[d_idx] += s_v[s_idx] * WIJ
        d_wf[d_idx] += s_w[s_idx] * WIJ

    def post_loop(self, d_uf, d_vf, d_wf, d_wij, d_idx,
                  d_ug, d_vg, d_wg, d_u, d_v, d_w):

        # calculation is done only for the relevant boundary particles.
        # d_wij (and d_uf) is 0 for particles sufficiently away from the
        # solid-fluid interface
        if d_wij[d_idx] > 1e-12:
            d_uf[d_idx] /= d_wij[d_idx]
            d_vf[d_idx] /= d_wij[d_idx]
            d_wf[d_idx] /= d_wij[d_idx]

        # Dummy velocities at the ghost points using Eq. (23),
        # d_u, d_v, d_w are the prescribed wall velocities.
        d_ug[d_idx] = 2*d_u[d_idx] - d_uf[d_idx]
        d_vg[d_idx] = 2*d_v[d_idx] - d_vf[d_idx]
        d_wg[d_idx] = 2*d_w[d_idx] - d_wf[d_idx]


class ContinuityEquation(Equation):
    r"""**Conservation of mass equation**

    Eq (6) in [Adami2012]:

    .. math::

        \frac{d\rho_a}{dt} = \rho_a \sum_b \frac{m_b}{\rho_b}
        \boldsymbol{v}_{ab} \cdot \nabla_a W_{ab}

    """

    def initialize(self, d_idx, d_arho):
        d_arho[d_idx] = 0.0

    def loop(self, d_idx, s_idx, d_arho, s_m, s_rho, d_rho, VIJ, DWIJ):
        vijdotdwij = VIJ[0] * DWIJ[0] + VIJ[1] * DWIJ[1] + VIJ[2] * DWIJ[2]
        d_arho[d_idx] += d_rho[d_idx] * vijdotdwij * s_m[s_idx] / s_rho[s_idx]


class ContinuitySolid(Equation):
    """Continuity equation for the solid's ghost particles.

    The key difference is that we use the ghost velocity ug, and not the
    particle velocity u.

    """
    def loop(self, d_idx, s_idx, d_rho, d_u, d_v, d_w, d_arho,
             s_m, s_rho, s_ug, s_vg, s_wg, DWIJ):
        Vj = s_m[s_idx] / s_rho[s_idx]
        rhoi = d_rho[d_idx]
        uij = d_u[d_idx] - s_ug[s_idx]
        vij = d_v[d_idx] - s_vg[s_idx]
        wij = d_w[d_idx] - s_wg[s_idx]
        vij_dot_dwij = uij*DWIJ[0] + vij*DWIJ[1] + wij*DWIJ[2]

        d_arho[d_idx] += rhoi*Vj*vij_dot_dwij


class StateEquation(Equation):
    r"""**Generalized Weakly Compressible Equation of State**

    .. math::

        p_a = p_0\left[ \left(\frac{\rho}{\rho_0}\right)^\gamma - b
        \right] + \mathcal{X}

    Notes
    -----
    This is the generalized Tait's equation of state and the suggested values
    in [Adami2013] are :math:`\mathcal{X} = 0`, :math:`\gamma=1` and
    :math:`b = 1`.

    The reference pressure :math:`p_0` is calculated from the artificial
    sound speed and reference density:

    .. math::

        p_0 = \frac{c^2\rho_0}{\gamma}
    """

    def __init__(self, dest, sources, p0, rho0, b=1.0):
        r"""
        Parameters
        ----------
        p0 : float
            reference pressure
        rho0 : float
            reference density
        b : float
            constant (default 1.0).
        """

        self.b = b
        self.p0 = p0
        self.rho0 = rho0
        super(StateEquation, self).__init__(dest, sources)

    def loop(self, d_idx, d_p, d_rho):
        d_p[d_idx] = self.p0 * (d_rho[d_idx]/self.rho0 - self.b)


class MomentumEquationPressureGradient(Equation):
    r"""**Momentum equation for the Transport Velocity Formulation: Pressure**

    Eq. (8) in [Adami2013]:

    .. math::

        \frac{d \boldsymbol{v}_a}{dt} = \frac{1}{m_a}\sum_b (V_a^2 +
        V_b^2)\left[-\bar{p}_{ab}\nabla_a W_{ab} \right]

    where

    .. math::

        \bar{p}_{ab} = \frac{\rho_b p_a + \rho_a p_b}{\rho_a + \rho_b}
    """

    def __init__(self, dest, sources, pb, gx=0., gy=0., gz=0.,
                 tdamp=0.0):

        r"""
        Parameters
        ----------
        pb : float
            background pressure
        gx : float
            Body force per unit mass along the x-axis
        gy : float
            Body force per unit mass along the y-axis
        gz : float
            Body force per unit mass along the z-axis
        tdamp : float
            damping time

        Notes
        -----
        This equation should have the destination as fluid and sources as
        fluid and boundary particles.

        This function also computes the contribution to the background
        pressure and accelerations due to a body force or gravity.

        The body forces are damped according to Eq. (13) in [Adami2012] to
        avoid instantaneous accelerations. By default, damping is neglected.
        """

        self.pb = pb
        self.gx = gx
        self.gy = gy
        self.gz = gz
        self.tdamp = tdamp
        super(MomentumEquationPressureGradient, self).__init__(dest, sources)

    def initialize(self, d_idx, d_au, d_av, d_aw, d_auhat, d_avhat, d_awhat):
        d_au[d_idx] = 0.0
        d_av[d_idx] = 0.0
        d_aw[d_idx] = 0.0

        d_auhat[d_idx] = 0.0
        d_avhat[d_idx] = 0.0
        d_awhat[d_idx] = 0.0

    def loop(self, d_idx, s_idx, d_m, d_rho, s_rho,
             d_au, d_av, d_aw, d_p, s_p,
             d_auhat, d_avhat, d_awhat, d_V, s_V, DWIJ):

        # averaged pressure Eq. (7)
        rhoi = d_rho[d_idx]
        rhoj = s_rho[s_idx]
        p_i = d_p[d_idx]
        p_j = s_p[s_idx]

        pij = rhoj * p_i + rhoi * p_j
        pij /= (rhoj + rhoi)

        # particle volumes; d_V is inverse volume
        Vi = 1./d_V[d_idx]
        Vj = 1./s_V[s_idx]
        Vi2 = Vi * Vi
        Vj2 = Vj * Vj

        # inverse mass of destination particle
        mi1 = 1.0/d_m[d_idx]

        # accelerations 1st term in Eq. (8)
        tmp = -pij * mi1 * (Vi2 + Vj2)

        d_au[d_idx] += tmp * DWIJ[0]
        d_av[d_idx] += tmp * DWIJ[1]
        d_aw[d_idx] += tmp * DWIJ[2]

        # contribution due to the background pressure Eq. (13)
        tmp = -self.pb * mi1 * (Vi2 + Vj2)

        d_auhat[d_idx] += tmp * DWIJ[0]
        d_avhat[d_idx] += tmp * DWIJ[1]
        d_awhat[d_idx] += tmp * DWIJ[2]

    def post_loop(self, d_idx, d_au, d_av, d_aw, t):
        # damped accelerations due to body or external force
        damping_factor = 1.0
        if t < self.tdamp:
            damping_factor = 0.5 * (sin((-0.5 + t/self.tdamp)*M_PI) + 1.0)

        d_au[d_idx] += self.gx * damping_factor
        d_av[d_idx] += self.gy * damping_factor
        d_aw[d_idx] += self.gz * damping_factor


class MomentumEquationViscosity(Equation):
    r"""**Momentum equation for the Transport Velocity Formulation: Viscosity**

    Eq. (8) in [Adami2013]:

    .. math::

           \frac{d \boldsymbol{v}_a}{dt} = \frac{1}{m_a}\sum_b (V_a^2 +
           V_b^2)\left[ \bar{\eta}_{ab}\hat{r}_{ab}\cdot \nabla_a W_{ab}
           \frac{\boldsymbol{v}_{ab}}{|\boldsymbol{r}_{ab}|}\right]

    where

    .. math::

        \bar{\eta}_{ab} = \frac{2\eta_a \eta_b}{\eta_a + \eta_b}
    """

    def __init__(self, dest, sources, nu):
        r"""
        Parameters
        ----------
        nu : float
            kinematic viscosity
        """

        self.nu = nu
        super(MomentumEquationViscosity, self).__init__(dest, sources)

    def initialize(self, d_idx, d_au, d_av, d_aw):
        d_au[d_idx] = 0.0
        d_av[d_idx] = 0.0
        d_aw[d_idx] = 0.0

    def loop(self, d_idx, s_idx, d_rho, s_rho, d_m, d_V, s_V,
             d_au, d_av, d_aw,
             R2IJ, EPS, DWIJ, VIJ, XIJ):

        # averaged shear viscosity Eq. (6)
        etai = self.nu * d_rho[d_idx]
        etaj = self.nu * s_rho[s_idx]

        etaij = 2 * (etai * etaj)/(etai + etaj)

        # scalar part of the kernel gradient
        Fij = DWIJ[0]*XIJ[0] + DWIJ[1]*XIJ[1] + DWIJ[2]*XIJ[2]

        # particle volumes, d_V is inverse volume.
        Vi = 1./d_V[d_idx]
        Vj = 1./s_V[s_idx]
        Vi2 = Vi * Vi
        Vj2 = Vj * Vj

        # accelerations 3rd term in Eq. (8)
        tmp = 1./d_m[d_idx] * (Vi2 + Vj2) * etaij * Fij/(R2IJ + EPS)

        d_au[d_idx] += tmp * VIJ[0]
        d_av[d_idx] += tmp * VIJ[1]
        d_aw[d_idx] += tmp * VIJ[2]


class MomentumEquationArtificialViscosity(Equation):
    r"""**Artificial viscosity for the momentum equation**

    Eq. (11) in [Adami2012]:

    .. math::

        \frac{d \boldsymbol{v}_a}{dt} = -\sum_b m_b \alpha h_{ab}
        c_{ab} \frac{\boldsymbol{v}_{ab}\cdot
        \boldsymbol{r}_{ab}}{\rho_{ab}\left(|r_{ab}|^2 + \epsilon
        \right)}\nabla_a W_{ab}

    where

    .. math::

        \rho_{ab} = \frac{\rho_a + \rho_b}{2}\\

        c_{ab} = \frac{c_a + c_b}{2}\\

        h_{ab} = \frac{h_a + h_b}{2}
    """
    def __init__(self, dest, sources, c0, alpha=0.1):
        r"""
        Parameters
        ----------
        alpha : float
            constant
        c0 : float
            speed of sound
        """

        self.alpha = alpha
        self.c0 = c0
        super(MomentumEquationArtificialViscosity, self).__init__(
            dest, sources
        )

    def initialize(self, d_idx, d_au, d_av, d_aw):
        d_au[d_idx] = 0.0
        d_av[d_idx] = 0.0
        d_aw[d_idx] = 0.0

    def loop(self, d_idx, s_idx, s_m, d_au, d_av, d_aw,
             RHOIJ1, R2IJ, EPS, DWIJ, VIJ, XIJ, HIJ):

        # v_{ab} \cdot r_{ab}
        vijdotrij = VIJ[0]*XIJ[0] + VIJ[1]*XIJ[1] + VIJ[2]*XIJ[2]

        # scalar part of the accelerations Eq. (11)
        piij = 0.0
        if vijdotrij < 0:
            muij = (HIJ * vijdotrij)/(R2IJ + EPS)

            piij = -self.alpha*self.c0*muij
            piij = s_m[s_idx] * piij*RHOIJ1

        d_au[d_idx] += -piij * DWIJ[0]
        d_av[d_idx] += -piij * DWIJ[1]
        d_aw[d_idx] += -piij * DWIJ[2]


class MomentumEquationArtificialStress(Equation):
    r"""**Artificial stress contribution to the Momentum Equation**

    .. math::

          \frac{d\boldsymbol{v}_a}{dt} = \frac{1}{m_a}\sum_b (V_a^2 +
          V_b^2)\left[ \frac{1}{2}(\boldsymbol{A}_a +
          \boldsymbol{A}_b) : \nabla_a W_{ab}\right]

    where the artificial stress terms are given by:

    .. math::

           \boldsymbol{A} = \rho \boldsymbol{v} (\tilde{\boldsymbol{v}}
         - \boldsymbol{v})

    """
    def initialize(self, d_idx, d_au, d_av, d_aw):
        d_au[d_idx] = 0.0
        d_av[d_idx] = 0.0
        d_aw[d_idx] = 0.0

    def loop(self, d_idx, s_idx, d_rho, d_u, d_v, d_w, d_V,
             d_uhat, d_vhat, d_what, d_au, d_av, d_aw, d_m,
             s_rho, s_u, s_v, s_w, s_V, s_uhat, s_vhat, s_what, DWIJ):
        rhoi = d_rho[d_idx]
        rhoj = s_rho[s_idx]

        # physical and advection velocities
        ui = d_u[d_idx]
        uhati = d_uhat[d_idx]
        vi = d_v[d_idx]
        vhati = d_vhat[d_idx]
        wi = d_w[d_idx]
        whati = d_what[d_idx]

        uj = s_u[s_idx]
        uhatj = s_uhat[s_idx]
        vj = s_v[s_idx]
        vhatj = s_vhat[s_idx]
        wj = s_w[s_idx]
        whatj = s_what[s_idx]

        # particle volumes; d_V is inverse volume.
        Vi = 1./d_V[d_idx]
        Vj = 1./s_V[s_idx]
        Vi2 = Vi * Vi
        Vj2 = Vj * Vj

        # artificial stress tensor
        Axxi = rhoi*ui*(uhati - ui)
        Axyi = rhoi*ui*(vhati - vi)
        Axzi = rhoi*ui*(whati - wi)
        Ayxi = rhoi*vi*(uhati - ui)
        Ayyi = rhoi*vi*(vhati - vi)
        Ayzi = rhoi*vi*(whati - wi)
        Azxi = rhoi*wi*(uhati - ui)
        Azyi = rhoi*wi*(vhati - vi)
        Azzi = rhoi*wi*(whati - wi)

        Axxj = rhoj*uj*(uhatj - uj)
        Axyj = rhoj*uj*(vhatj - vj)
        Axzj = rhoj*uj*(whatj - wj)
        Ayxj = rhoj*vj*(uhatj - uj)
        Ayyj = rhoj*vj*(vhatj - vj)
        Ayzj = rhoj*vj*(whatj - wj)
        Azxj = rhoj*wj*(uhatj - uj)
        Azyj = rhoj*wj*(vhatj - vj)
        Azzj = rhoj*wj*(whatj - wj)

        # contraction of stress tensor with kernel gradient
        Ax = 0.5*(
            (Axxi + Axxj)*DWIJ[0] +
            (Axyi + Axyj)*DWIJ[1] +
            (Axzi + Axzj)*DWIJ[2]
        )

        Ay = 0.5*(
            (Ayxi + Ayxj)*DWIJ[0] +
            (Ayyi + Ayyj)*DWIJ[1] +
            (Ayzi + Ayzj)*DWIJ[2]
        )

        Az = 0.5*(
            (Azxi + Azxj)*DWIJ[0] +
            (Azyi + Azyj)*DWIJ[1] +
            (Azzi + Azzj)*DWIJ[2]
        )

        # accelerations 2nd part of Eq. (8)
        tmp = 1./d_m[d_idx] * (Vi2 + Vj2)

        d_au[d_idx] += tmp * Ax
        d_av[d_idx] += tmp * Ay
        d_aw[d_idx] += tmp * Az


class SolidWallNoSlipBC(Equation):
    r"""**Solid wall boundary condition** [Adami2012]_

    This boundary condition is to be used with fixed ghost particles
    in SPH simulations and is formulated for the general case of
    moving boundaries.

    The velocity and pressure of the fluid particles is extrapolated
    to the ghost particles and these values are used in the equations
    of motion.

    No-penetration:

    Ghost particles participate in the continuity and state equations
    with fluid particles. This means as fluid particles approach the
    wall, the pressure of the ghost particles increases to generate a
    repulsion force that prevents particle penetration.

    No-slip:

    Extrapolation is used to set the `dummy` velocity of the ghost
    particles for viscous interaction. First, the smoothed velocity
    field of the fluid phase is extrapolated to the wall particles:

    .. math::

        \tilde{v}_a = \frac{\sum_b v_b W_{ab}}{\sum_b W_{ab}}

    In the second step, for the viscous interaction in Eqs. (10) in [Adami2012]
    and Eq. (8) in [Adami2013], the velocity of the ghost particles is
    assigned as:

    .. math::

       v_b = 2v_w -\tilde{v}_a,

    where :math:`v_w` is the prescribed wall velocity and :math:`v_b`
    is the ghost particle in the interaction.
    """

    def __init__(self, dest, sources, nu):
        r"""
        Parameters
        ----------
        nu : float
            kinematic viscosity

        Notes
        -----
        For this equation the destination particle array should be the
        fluid and the source should be ghost or boundary particles. The
        boundary particles must define a prescribed velocity :math:`u_0,
        v_0, w_0`
        """

        self.nu = nu
        super(SolidWallNoSlipBC, self).__init__(dest, sources)

    def initialize(self, d_idx, d_au, d_av, d_aw):
        d_au[d_idx] = 0.0
        d_av[d_idx] = 0.0
        d_aw[d_idx] = 0.0

    def loop(self, d_idx, s_idx, d_m, d_rho, s_rho, d_V, s_V,
             d_u, d_v, d_w,
             d_au, d_av, d_aw,
             s_ug, s_vg, s_wg,
             DWIJ, R2IJ, EPS, XIJ):

        # averaged shear viscosity Eq. (6).
        etai = self.nu * d_rho[d_idx]
        etaj = self.nu * s_rho[s_idx]

        etaij = 2 * (etai * etaj)/(etai + etaj)

        # particle volumes; d_V inverse volume.
        Vi = 1./d_V[d_idx]
        Vj = 1./s_V[s_idx]
        Vi2 = Vi * Vi
        Vj2 = Vj * Vj

        # scalar part of the kernel gradient
        Fij = XIJ[0]*DWIJ[0] + XIJ[1]*DWIJ[1] + XIJ[2]*DWIJ[2]

        # viscous contribution (third term) from Eq. (8), with VIJ
        # defined appropriately using the ghost values
        tmp = 1./d_m[d_idx] * (Vi2 + Vj2) * (etaij * Fij/(R2IJ + EPS))

        d_au[d_idx] += tmp * (d_u[d_idx] - s_ug[s_idx])
        d_av[d_idx] += tmp * (d_v[d_idx] - s_vg[s_idx])
        d_aw[d_idx] += tmp * (d_w[d_idx] - s_wg[s_idx])


class SolidWallPressureBC(Equation):
    r"""**Solid wall pressure boundary condition** [Adami2012]_

    This boundary condition is to be used with fixed ghost particles
    in SPH simulations and is formulated for the general case of
    moving boundaries.

    The velocity and pressure of the fluid particles is extrapolated
    to the ghost particles and these values are used in the equations
    of motion.

    Pressure boundary condition:

    The pressure of the ghost particle is also calculated from the
    fluid particle by interpolation using:

    .. math::

        p_g = \frac{\sum_f p_f W_{gf} + \boldsymbol{g - a_g} \cdot
        \sum_f \rho_f \boldsymbol{r}_{gf}W_{gf}}{\sum_f W_{gf}},

    where the subscripts `g` and `f` relate to the ghost and fluid
    particles respectively.

    Density of the wall particle is then set using this pressure

    .. math::

        \rho_w=\rho_0\left(\frac{p_w - \mathcal{X}}{p_0} +
        1\right)^{\frac{1}{\gamma}}
    """

    def __init__(self, dest, sources, rho0, p0, b=1.0, gx=0.0, gy=0.0, gz=0.0):
        r"""
        Parameters
        ----------
        rho0 : float
            reference density
        p0 : float
            reference pressure
        b : float
            constant (default 1.0)
        gx : float
            Body force per unit mass along the x-axis
        gy : float
            Body force per unit mass along the y-axis
        gz : float
            Body force per unit mass along the z-axis

        Notes
        -----
        For a two fluid system (boundary, fluid), this equation must be
        instantiated with boundary as the destination and fluid as the
        source.

        The boundary particle array must additionally define a property
        :math:`wij` for the denominator in Eq. (27) from [Adami2012]. This
        array sums the kernel terms from the ghost particle to the fluid
        particle.
        """

        self.rho0 = rho0
        self.p0 = p0
        self.b = b
        self.gx = gx
        self.gy = gy
        self.gz = gz

        super(SolidWallPressureBC, self).__init__(dest, sources)

    def initialize(self, d_idx, d_p, d_wij):
        d_p[d_idx] = 0.0
        d_wij[d_idx] = 0.0

    def loop(self, d_idx, s_idx, d_p, s_p, d_wij, s_rho,
             d_au, d_av, d_aw, WIJ, XIJ):

        # numerator of Eq. (27) ax, ay and az are the prescribed wall
        # accelerations which must be defined for the wall boundary
        # particle
        gdotxij = (self.gx - d_au[d_idx])*XIJ[0] + \
            (self.gy - d_av[d_idx])*XIJ[1] + \
            (self.gz - d_aw[d_idx])*XIJ[2]

        d_p[d_idx] += s_p[s_idx]*WIJ + s_rho[s_idx]*gdotxij*WIJ

        # denominator of Eq. (27)
        d_wij[d_idx] += WIJ

    def post_loop(self, d_idx, d_wij, d_p, d_rho):
        # extrapolated pressure at the ghost particle
        if d_wij[d_idx] > 1e-14:
            d_p[d_idx] /= d_wij[d_idx]

        # update the density from the pressure Eq. (28)
        d_rho[d_idx] = self.rho0 * (d_p[d_idx]/self.p0 + self.b)



"""Three-dimensional dam break over a dry bed. (14 hours)

The case is described as a SPHERIC benchmark
https://app.spheric-sph.org/sites/spheric/files/SPHERIC_Test2_v1p1.pdf

By default the simulation runs for 6 seconds of simulation time.
"""

import numpy as np

from pysph.base.kernels import WendlandQuintic
from pysph.examples._db_geometry import DamBreak3DGeometry
from pysph.solver.application import Application
from pysph.sph.integrator import EPECIntegrator
from pysph.sph.scheme import WCSPHScheme

dim = 3

dt = 1e-5
tf = 6.0

# parameter to change the resolution
dx = 0.02
nboundary_layers = 1
hdx = 1.3
ro = 1000.0
h0 = dx * hdx
gamma = 7.0
alpha = 0.25
beta = 0.0
c0 = 10.0 * np.sqrt(2.0 * 9.81 * 0.55)


class DamBreak3D(Application):
    def add_user_options(self, group):
        group.add_argument(
            '--dx', action='store', type=float, dest='dx',  default=dx,
            help='Particle spacing.'
        )
        group.add_argument(
            '--hdx', action='store', type=float, dest='hdx',  default=hdx,
            help='Specify the hdx factor where h = hdx * dx.'
        )

    def consume_user_options(self):
        dx = self.options.dx
        self.dx = dx
        self.hdx = self.options.hdx
        self.geom = DamBreak3DGeometry(
            dx=dx, nboundary_layers=nboundary_layers, hdx=self.hdx, rho0=ro
        )
        self.co = 10.0 * self.geom.get_max_speed(g=9.81)

    def create_scheme(self):
        s = WCSPHScheme(
            ['fluid'], ['boundary', 'obstacle'], dim=dim, rho0=ro, c0=c0,
            h0=h0, hdx=hdx, gz=-9.81, alpha=alpha, beta=beta, gamma=gamma,
            hg_correction=True, tensile_correction=False
        )
        return s

    def configure_scheme(self):
        s = self.scheme
        hdx = self.hdx
        kernel = WendlandQuintic(dim=dim)
        h0 = self.dx * hdx
        s.configure(h0=h0, hdx=hdx)
        dt = 0.25*h0/(1.1 * self.co)
        s.configure_solver(
            kernel=kernel, integrator_cls=EPECIntegrator, tf=tf, dt=dt,
            adaptive_timestep=True, n_damp=50,
            output_at_times=[0.4, 0.6, 1.0]
        )

    def create_particles(self):
        return self.geom.create_particles()

    def customize_output(self):
        self._mayavi_config('''
        viewer.scalar = 'u'
        b = particle_arrays['boundary']
        b.plot.actor.mapper.scalar_visibility = False
        b.plot.actor.property.opacity = 0.1
        ''')

    def post_process(self, info_fname):
        if self.rank > 0:
            return
        import os
        if os.path.exists(info_fname):
            self.read_info(info_fname)
        else:
            return

        if len(self.output_files) == 0:
            return

        try:
            import matplotlib
            matplotlib.use('Agg')
            import matplotlib.pyplot as plt
        except ImportError:
            print("Post processing requires matplotlib")
            return
        from pysph.solver.utils import iter_output
        from pysph.examples import db_exp_data as dbd
        from pysph.tools.interpolator import Interpolator
        H = self.geom.fluid_column_height
        factor_y = 1/(ro*9.81*H)
        factor_x = np.sqrt(9.81/H)

        t1, t3, data_p1, data_p3 = dbd.get_kleefsman_data()
        files = self.output_files

        t = []
        p0 = []

        x_probe = self.geom.obstacle_center_x - self.geom.obstacle_length*0.5
        p_x = np.repeat(x_probe, 2)
        p_y = np.repeat(0, 2)
        p_z = np.array([0.021, 0.101])

        for sd, arrays1, arrays2, arrays3 in iter_output(
                files, "fluid", "obstacle", "boundary"
        ):
            t.append(sd["t"]*factor_x)
            interp = Interpolator([arrays1, arrays2, arrays3],
                                  x=p_x, y=p_y, z=p_z, method="shepard")
            p0.append(interp.interpolate('p')*factor_y)

        fname = os.path.join(self.output_dir, 'results.npz')
        t, p0 = list(map(np.asarray, (t, p0)))
        np.savez(fname, t=t, p0=p0)

        p1 = p0[:, 0]
        p3 = p0[:, 1]

        fig1 = plt.figure()
        plt.plot(t, p1, label="p1 computed", figure=fig1)
        plt.plot(t1, data_p1, label="Kleefsman et al.", figure=fig1)
        plt.legend()
        plt.ylabel(r"$\frac{P}{\rho gH}$")
        plt.xlabel(r"$t \sqrt{\frac{g}{H}} $")
        plt.title("P1")
        plt.savefig(os.path.join(self.output_dir, 'p1_vs_t.png'))

        fig2 = plt.figure()
        plt.plot(t, p3, label="p3 computed", figure=fig2)
        plt.plot(t3, data_p3, label="Kleefsman et al.", figure=fig2)
        plt.legend()
        plt.ylabel(r"$\frac{P}{\rho gH}$")
        plt.xlabel(r"$t \sqrt{\frac{g}{H}} $")
        plt.title("P3")
        plt.savefig(os.path.join(self.output_dir, 'p3_vs_t.png'))


if __name__ == '__main__':
    app = DamBreak3D()
    app.run()
    app.post_process(app.info_filename)



# cython: language_level=3, embedsignature=True
# distutils: language=c++
from .nnps_base cimport *
from libcpp.vector cimport vector
cimport cython

import numpy as np
cimport numpy as np

ctypedef unsigned int u_int

cdef extern from 'math.h':
    int abs(int) nogil
    double ceil(double) nogil
    double floor(double) nogil
    double fabs(double) nogil
    double fmax(double, double) nogil
    double fmin(double, double) nogil

cdef extern from "<algorithm>" namespace "std" nogil:
    OutputIter copy[InputIter,OutputIter](InputIter,InputIter,OutputIter)

cdef struct cOctreeNode:
    bint is_leaf
    double length
    double xmin[3]
    double hmax
    int num_particles
    int level

    int start_index
    cOctreeNode* children[8]
    cOctreeNode* parent

cdef class OctreeNode:
    ##########################################################################
    # Data Attributes
    ##########################################################################
    cdef cOctreeNode* _node

    cdef public bint is_leaf
    cdef public double length
    cdef public np.ndarray xmin
    cdef public double hmax
    cdef public int num_particles
    cdef public int level

    ##########################################################################
    # Member functions
    ##########################################################################

    cdef void wrap_node(self, cOctreeNode* node)

    cpdef OctreeNode get_parent(self)

    cpdef UIntArray get_indices(self, Octree tree)

    cpdef list get_children(self)

    cpdef plot(self, ax, color = *)

cdef class Octree:
    ##########################################################################
    # Data Attributes
    ##########################################################################
    cdef cOctreeNode* root
    cdef vector[cOctreeNode*]* leaf_cells
    cdef u_int* pids

    cdef int _next_pid
    cdef public int num_particles

    cdef public int leaf_max_particles
    cdef public double hmax
    cdef public double length
    cdef public int depth
    cdef public bint test_parallel
    cdef double machine_eps

    cdef double xmin[3]
    cdef double xmax[3]

    ##########################################################################
    # Member functions
    ##########################################################################

    cdef inline double _get_eps(self, double length, double* xmin) nogil

    cdef inline void _calculate_domain(self, NNPSParticleArrayWrapper pa)

    cdef inline cOctreeNode* _new_node(self, double* xmin, double length,
            double hmax = *, int level = *, cOctreeNode* parent = *,
            int num_particles = *, bint is_leaf = *) nogil

    cdef inline void _delete_tree(self, cOctreeNode* node)

    cdef int _c_build_tree(self, NNPSParticleArrayWrapper pa,
            vector[u_int]* indices, double* xmin, double length,
            cOctreeNode* node, int level) nogil

    cdef int _c_build_tree_level1(self, NNPSParticleArrayWrapper pa,
             double* xmin, double length, cOctreeNode* node, int num_threads) nogil    
    
    cdef int _c_build_tree_bfs(self, NNPSParticleArrayWrapper pa,
            u_int* p_indices, vector[cOctreeNode *]* level_nodes,
            int level, int num_threads) nogil

    cdef void _plot_tree(self, OctreeNode node, ax)

    cdef int c_build_tree(self, NNPSParticleArrayWrapper pa_wrapper, bint test_parallel = *)

    cdef void _c_get_leaf_cells(self, cOctreeNode* node)

    cdef void c_get_leaf_cells(self)

    cdef cOctreeNode* c_find_point(self, double x, double y, double z)

    cpdef int build_tree(self, ParticleArray pa, bint test_parallel = *)

    cpdef delete_tree(self)

    cpdef OctreeNode get_root(self)

    cpdef list get_leaf_cells(self)

    cpdef OctreeNode find_point(self, double x, double y, double z)

    cpdef plot(self, ax)

cdef class CompressedOctree(Octree):
    ##########################################################################
    # Data Attributes
    ##########################################################################
    cdef double dbl_max

    ##########################################################################
    # Member functions
    ##########################################################################

    cdef int _c_build_tree(self, NNPSParticleArrayWrapper pa,
            vector[u_int]* indices, double* xmin, double length,
            cOctreeNode* node, int level) nogil
    
    cdef int c_build_tree(self, NNPSParticleArrayWrapper pa_wrapper, bint test_parallel = *)
    
    



"""Ten sphere of different density falling into a hydrostatic tank
implemented using Akinci in 2d. (15 minutes)

Check basic equations of SPH to throw a ball inside the vessel

Geometry will be like




    ^  ||        ___      ___                            ___      ___       ||
    |  ||       /   \    /   \                          /   \    /   \      ||
    |  ||       |   |    |   |                          |   |    |   |      ||
    |  ||    ___\___/ ___\___/  ___                 ___ \___/ ___\___/ ___  ||
    |  ||   /   \    /   \     /   \               /   \     /   \    /   \ ||
    |  ||   |   |    |   |     |   |               |   |     |   |    |   | ||
    |  ||   \___/    \___/     \___/               \___/     \___/    \___/ ||
    |  ||___________________________________________________________________||
500mm  ||                          ^                                        ||
    |  ||                          |                                        ||
    |  ||                          |                                        ||
    |  ||                          |300 mm                                  ||
    |  ||                          |                                        ||
    |  ||                          |                                        ||
    |  ||                          |                                        ||
    v  ||__________________________v________________________________________||
       ||___________________________________________________________________||

       <--------------------------1000mm------------------------------------->






"""
from __future__ import print_function
import numpy as np

# PySPH base and carray imports
from pysph.base.utils import (get_particle_array_wcsph,
                              get_particle_array_rigid_body)
from pysph.base.kernels import CubicSpline

from pysph.solver.solver import Solver
from pysph.sph.integrator import EPECIntegrator
from pysph.sph.integrator_step import WCSPHStep

from pysph.sph.equation import Group
from pysph.sph.basic_equations import (XSPHCorrection, SummationDensity)
from pysph.sph.wc.basic import TaitEOSHGCorrection, MomentumEquation
from pysph.solver.application import Application
from pysph.sph.rigid_body import (
    BodyForce, RigidBodyCollision, SummationDensityBoundary, RigidBodyMoments,
    RigidBodyMotion, AkinciRigidFluidCoupling, RK2StepRigidBody)


def get_2d_dam(length=10, height=15, dx=0.1, layers=2):
    _x = np.arange(0, length, dx)
    _y = np.arange(0, height, dx)

    x, y = np.meshgrid(_x, _y)
    x, y = x.ravel(), y.ravel()

    # get particles inside the tank
    cond = ((x > (layers - 1) * dx)) & ((x < (x[-1] - (layers - 1) * dx)) &
                                        (y > (layers - 1) * dx))

    # exclude inside particles
    x, y = x[~cond], y[~cond]

    return x, y


def get_2d_block(length=10, height=15, dx=0.1):
    x = np.arange(0, length, dx)
    y = np.arange(0, height, dx)

    x, y = np.meshgrid(x, y)
    x, y = x.ravel(), y.ravel()
    return x, y


def get_fluid_and_dam_geometry(d_l, d_h, f_l, f_h, d_layers, d_dx, f_dx,
                               fluid_left_extreme=None):
    xd, yd = get_2d_dam(d_l, d_h, d_dx, d_layers)
    xf, yf = get_2d_block(f_l, f_h, f_dx)

    if fluid_left_extreme:
        x_trans, y_trans = fluid_left_extreme
        xf += x_trans
        yf += y_trans

    else:
        xf += 2 * d_dx
        yf += 2 * d_dx

    return xd, yd, xf, yf


def get_circle(centre=[0, 0], radius=1, dx=0.1):
    x = np.arange(0, radius * 2, dx)
    y = np.arange(0, radius * 2, dx)

    x, y = np.meshgrid(x, y)
    x, y = x.ravel(), y.ravel()

    cond = ((x - radius)**2 + (y - radius)**2) <= radius**2

    x, y = x[cond], y[cond]

    x_trans = centre[0] - radius
    y_trans = centre[1] - radius

    x = x + x_trans
    y = y + y_trans

    return x, y


def create_ten_circles(radius=20 * 1e-3, spacing=1 * 1e-3,
                       fluid_height=300 * 1e-3):
    x1, y1 = get_circle(centre=[100 * 1e-3, fluid_height + radius + 30 * 1e-3],
                        radius=radius, dx=spacing)

    x2, y2 = x1 + 2 * radius, y1 + 3 * radius

    x3, y3 = x2 + 2 * radius, y1

    x4, y4 = x3 + 2 * radius, y2

    x5, y5 = x4 + 2 * radius, y3

    x_left = np.concatenate([x1, x2, x3, x4, x5])
    y_left = np.concatenate([y1, y2, y3, y4, y5])

    # x_middle, y_middle = x1 + 400 * 1e-3, y1 + 300 * 1e-3

    x_right = x_left + 500 * 1e-3
    y_right = y_left

    x = np.concatenate([x_left, x_right])
    y = np.concatenate([y_left, y_right])
    return x, y


def get_rho_of_each_sphere(xc, yc, radius=20 * 1e-3, spacing=1 * 1e-3):
    x1, y1 = get_circle(radius=radius, dx=spacing)
    pars = len(x1)

    rho = np.ones_like(xc)
    no_of_spheres = int(len(rho) / len(x1))

    for i in range(no_of_spheres):
        if i < 5:
            rho[i * pars:(i + 1) * pars] = 500
        if i >= 5:
            rho[i * pars:(i + 1) * pars] = 1500

    return rho


def get_body_id_of_each_sphere(xc, yc, radius=20 * 1e-3, spacing=1 * 1e-3):
    x1, y1 = get_circle(radius=radius, dx=spacing)
    pars = len(x1)

    body_id = np.ones_like(xc, dtype=int)
    no_of_spheres = int(len(body_id) / len(x1))

    for i in range(no_of_spheres):
        body_id[i * pars:(i + 1) * pars] = i

    return body_id


class RigidFluidCoupling(Application):
    def initialize(self):
        self.dam_length = 1000 * 1e-3
        self.dam_height = 500 * 1e-3
        self.dam_spacing = 2 * 1e-3
        self.dam_layers = 3

        self.fluid_length = (
            1000 * 1e-3 - 3 * self.dam_layers * self.dam_spacing)
        self.fluid_height = 300 * 1e-3
        self.fluid_spacing = 5 * 1e-3
        self.fluid_rho = 1000.

        self.sphere_radius = 30 * 1e-3
        self.sphere_spacing = 4 * 1e-3

        # simulation properties
        self.hdx = 1.2
        self.co = 2 * np.sqrt(2 * 9.81 * self.fluid_height)
        self.alpha = 0.1

    def create_particles(self):
        # get the geometry
        xt, yt, xf, yf = get_fluid_and_dam_geometry(
            self.dam_length, self.dam_height, self.fluid_length,
            self.fluid_height, self.dam_layers, self.dam_spacing,
            self.fluid_spacing, [3 * self.dam_spacing, 3 * self.dam_spacing])

        # create fluid particle array
        m = self.fluid_rho * self.fluid_spacing * self.fluid_spacing
        rho = self.fluid_rho
        h = self.hdx * self.fluid_spacing
        fluid = get_particle_array_wcsph(x=xf, y=yf, h=h, m=m, rho=rho,
                                         name="fluid")

        # create tank particle array
        m = self.fluid_rho * self.dam_spacing * self.dam_spacing
        rho = 1000
        rad_s = self.dam_spacing / 2.
        h = self.hdx * self.dam_spacing
        V = self.dam_spacing**2
        tank = get_particle_array_wcsph(x=xt, y=yt, h=h, m=m, rho=rho,
                                        rad_s=rad_s, V=V, name="tank")
        for name in ['fx', 'fy', 'fz']:
            tank.add_property(name)

        xc, yc = create_ten_circles(radius=self.sphere_radius,
                                    spacing=self.sphere_spacing,
                                    fluid_height=self.fluid_height)

        # get density of each sphere
        rho = get_rho_of_each_sphere(xc, yc, radius=self.sphere_radius,
                                     spacing=self.sphere_spacing)
        # get bodyid for each sphere
        body_id = get_body_id_of_each_sphere(xc, yc, radius=self.sphere_radius,
                                             spacing=self.sphere_spacing)
        m = rho * self.sphere_spacing**2
        h = self.hdx * self.sphere_spacing
        rad_s = self.sphere_spacing / 2.
        V = self.sphere_spacing**2
        cs = 0.0
        cube = get_particle_array_rigid_body(x=xc, y=yc, h=h, m=m, rho=rho,
                                             rad_s=rad_s, V=V, cs=cs,
                                             body_id=body_id, name="cube")
        return [fluid, tank, cube]

    def create_solver(self):
        kernel = CubicSpline(dim=2)

        integrator = EPECIntegrator(fluid=WCSPHStep(), cube=RK2StepRigidBody(),
                                    tank=WCSPHStep())

        dt = 1 * 1e-4
        print("DT: %s" % dt)
        tf = 1
        solver = Solver(
            kernel=kernel,
            dim=2,
            integrator=integrator,
            dt=dt,
            tf=tf,
            adaptive_timestep=False, )

        return solver

    def create_equations(self):
        equations = [
            Group(equations=[
                BodyForce(dest='cube', sources=None, gy=-9.81),
            ], real=False),
            Group(equations=[
                SummationDensity(
                    dest='fluid',
                    sources=['fluid'], ),
                SummationDensityBoundary(
                    dest='fluid', sources=['tank', 'cube'], fluid_rho=1000.0)
            ]),

            # Tait equation of state
            Group(equations=[
                TaitEOSHGCorrection(dest='fluid', sources=None,
                                    rho0=self.fluid_rho, c0=self.co,
                                    gamma=7.0),
            ], real=False),
            Group(equations=[
                MomentumEquation(dest='fluid', sources=['fluid'],
                                 alpha=self.alpha, beta=0.0, c0=self.co,
                                 gy=-9.81),
                AkinciRigidFluidCoupling(dest='fluid',
                                         sources=['cube', 'tank']),
                XSPHCorrection(dest='fluid', sources=['fluid', 'tank',
                                                      'cube']),
            ]),
            Group(equations=[
                RigidBodyCollision(dest='cube', sources=['tank', 'cube'],
                                   kn=1e5)
            ]),
            Group(equations=[RigidBodyMoments(dest='cube', sources=None)]),
            Group(equations=[RigidBodyMotion(dest='cube', sources=None)]),
        ]
        return equations

    def geometry(self):
        import matplotlib.pyplot as plt
        # please run this function to know how
        # geometry looks like
        x_tank, y_tank, x_fluid, y_fluid = get_fluid_and_dam_geometry(
            self.dam_length, self.dam_height, self.fluid_length,
            self.fluid_height, self.dam_layers, self.dam_spacing,
            self.fluid_spacing, [3 * self.dam_spacing, 3 * self.dam_spacing])

        x_cube, y_cube = create_ten_circles(radius=self.sphere_radius,
                                            spacing=self.sphere_spacing,
                                            fluid_height=self.fluid_height)
        plt.scatter(x_fluid, y_fluid)
        plt.scatter(x_tank, y_tank)
        plt.scatter(x_cube, y_cube)
        plt.axes().set_aspect('equal', 'datalim')
        print("done")
        plt.show()


if __name__ == '__main__':
    app = RigidFluidCoupling()
    app.run()
    # app.geometry()



import numpy as np
from pysph.base.utils import get_particle_array
from pysph.solver.application import Application
from pysph.sph.scheme import WCSPHScheme


class EllipticalDrop(Application):
    def create_particles(self):
        dx = 0.025
        x, y = np.mgrid[-1.05:1.05:dx, -1.05:1.05:dx]
        mask = x*x + y*y < 1
        x = x[mask]
        y = y[mask]
        rho = 1.0
        h = 1.3*dx
        m = rho*dx*dx
        pa = get_particle_array(
            name='fluid', x=x, y=y, u=-100*x, v=100*y, rho=rho,
            m=m, h=h
        )
        self.scheme.setup_properties([pa])
        return [pa]

    def create_scheme(self):
        s = WCSPHScheme(
            ['fluid'], [], dim=2, rho0=1.0, c0=1400,
            h0=1.3*0.025, hdx=1.3, gamma=7.0, alpha=0.1, beta=0.0
        )
        dt = 5e-6
        tf = 0.0076
        s.configure_solver(
            dt=dt, tf=tf,
        )
        return s


if __name__ == '__main__':
    app = EllipticalDrop(fname='ed')
    app.run()



# cython: language_level=3, embedsignature=True
# distutils: language=c++
# numpy
cimport numpy as np
cimport cython

from libcpp.map cimport map
from libcpp.vector cimport vector

# PyZoltan CArrays
from cyarray.carray cimport UIntArray, IntArray, DoubleArray, LongArray

# local imports
from .particle_array cimport ParticleArray
from .point cimport *

cdef extern from 'math.h':
    int abs(int) nogil
    double ceil(double) nogil
    double floor(double) nogil
    double fabs(double) nogil
    double fmax(double, double) nogil
    double fmin(double, double) nogil

cdef extern from 'limits.h':
    cdef unsigned int UINT_MAX
    cdef int INT_MAX

# ZOLTAN ID TYPE AND PTR
ctypedef unsigned int ZOLTAN_ID_TYPE
ctypedef unsigned int* ZOLTAN_ID_PTR

cdef inline double norm2(double x, double y, double z) nogil:
    return x*x + y*y + z*z

@cython.cdivision(True)
cdef inline int real_to_int(double real_val, double step) nogil:
    """ Return the bin index to which the given position belongs.

    Parameters
    ----------
    val -- The coordinate location to bin
    step -- the bin size

    Examples
    --------
    >>> real_to_int(1.5, 1.0)
    1
    >>> real_to_int(-0.5, 1.0)
    -1
    """
    cdef int ret_val = <int>floor( real_val/step )

    return ret_val

cdef inline void find_cell_id_raw(double x, double y, double z, double
                           cell_size, int *ix, int *iy, int *iz) nogil:
    """ Find the cell index for the corresponding point

    Parameters
    ----------
    x, y, z: double
        the point for which the index is sought
    cell_size : double
        the cell size to use
    ix, iy, iz : int*
        output parameter holding the cell index

    Notes
    ------
    Performs a box sort based on the point and cell size

    Uses the function  `real_to_int`

    """
    ix[0] = real_to_int(x, cell_size)
    iy[0] = real_to_int(y, cell_size)
    iz[0] = real_to_int(z, cell_size)

@cython.boundscheck(False)
@cython.wraparound(False)
cdef inline long flatten_raw(int x, int y, int z, int* ncells_per_dim,
        int dim) nogil:
    """Return a flattened index for a cell

    The flattening is determined using the row-order indexing commonly
    employed in SPH. This would need to be changed for hash functions
    based on alternate orderings.

    """
    cdef long ncx = ncells_per_dim[0]
    cdef long ncy = ncells_per_dim[1]

    return <long>( x + ncx * y + ncx*ncy * z )


@cython.boundscheck(False)
@cython.wraparound(False)
cdef inline long flatten(cIntPoint cid, IntArray ncells_per_dim, int dim) nogil:
    """Return a flattened index for a cell

    The flattening is determined using the row-order indexing commonly
    employed in SPH. This would need to be changed for hash functions
    based on alternate orderings.

    """
    return flatten_raw(cid.x, cid.y, cid.z, ncells_per_dim.data, dim)

@cython.boundscheck(False)
@cython.wraparound(False)
cdef inline long get_valid_cell_index(int cid_x, int cid_y, int cid_z,
        int* ncells_per_dim, int dim, int n_cells) nogil:
    """Return the flattened index for a valid cell"""
    cdef long ncx = ncells_per_dim[0]
    cdef long ncy = ncells_per_dim[1]
    cdef long ncz = ncells_per_dim[2]

    cdef long cell_index = -1

    # basic test for valid indices. Since we bin the particles with
    # respect to the origin, negative indices can never occur.
    cdef bint is_valid = (
        (ncx > cid_x > -1) and (ncy > cid_y > -1) and (ncz > cid_z > -1)
    )

    # Given the validity of the cells, return the flattened cell index
    if is_valid:
        cell_index = flatten_raw(cid_x, cid_y, cid_z, ncells_per_dim, dim)

        if not (-1 < cell_index < n_cells):
            cell_index = -1

    return cell_index

cdef cIntPoint find_cell_id(cPoint pnt, double cell_size)

cpdef UIntArray arange_uint(int start, int stop=*)

cpdef int get_number_of_threads()

# Basic particle array wrapper used for NNPS
cdef class NNPSParticleArrayWrapper:
    cdef public DoubleArray x,y,z,h
    cdef public UIntArray gid
    cdef public IntArray tag
    cdef public ParticleArray pa

    cdef public str name
    cdef public int np

    # get the number of particles
    cdef int get_number_of_particles(self)


cdef class DomainManager:
    cdef public object backend
    cdef public object manager


cdef class DomainManagerBase:
    cdef public double xmin, xmax
    cdef public double ymin, ymax
    cdef public double zmin, zmax

    cdef public double xtranslate
    cdef public double ytranslate
    cdef public double ztranslate

    cdef public int dim
    cdef public bint periodic_in_x, periodic_in_y, periodic_in_z
    cdef public bint is_periodic
    cdef public bint mirror_in_x, mirror_in_y, mirror_in_z
    cdef public bint is_mirror

    cdef public object props
    cdef public list copy_props
    cdef public list pa_wrappers     # NNPS particle array wrappers
    cdef public int narrays          # number of arrays
    cdef public double cell_size     # distance to create ghosts
    cdef public double hmin          # minimum h
    cdef public bint in_parallel     # Flag to determine if in parallel
    cdef public double radius_scale  # Radius scale for kernel
    cdef public double n_layers      # Number of layers of ghost particles

    #cdef double dbl_max              # Maximum value of double

    # remove ghost particles from a previous iteration
    cpdef _remove_ghosts(self)


# Domain limits for the simulation
cdef class CPUDomainManager(DomainManagerBase):
    cdef public bint use_double
    cdef public object dtype
    cdef public double dtype_max
    cdef public list ghosts

    # box-wrap particles within the physical domain
    cdef _box_wrap_periodic(self)

    # Convenience function to add a value to a carray
    cdef _add_to_array(self, DoubleArray arr, double disp, int start=*)

    # Convenience function to multiply a value to a carray
    cdef _mul_to_array(self, DoubleArray arr, double val)

    # Convenience function to add a carray to a carray elementwise
    cdef _add_array_to_array(self, DoubleArray arr, DoubleArray translate)

    # Convenience function to add a value to a carray
    cdef _change_velocity(self, DoubleArray arr, double disp)

    # create new periodic ghosts
    cdef _create_ghosts_periodic(self)

    # create new mirror ghosts
    cdef _create_ghosts_mirror(self)

    # Compute the cell size across processors. The cell size is taken
    # as max(h)*radius_scale
    cdef _compute_cell_size_for_binning(self)


# Cell to hold particle indices
cdef class Cell:
    ############################################################################
    # Data Attributes
    ############################################################################
    cdef cIntPoint _cid                 # Spatial index for the cell
    cdef public bint is_boundary        # Flag to indicate boundary cells
    cdef int narrays                    # Number of particle arrays
    cdef public list lindices           # Local indices for particles
    cdef public list gindices           # Global indices for binned particles
    cdef list nparticles                # Number of particles in the cell
    cdef double cell_size               # bin size
    cdef public cPoint centroid         # Centroid computed from indices
    cdef cPoint boxmin                  # Bounding box min for the cell
    cdef cPoint boxmax                  # Bounding box max for the cell
    cdef int layers                     # Layers to compute bounding box
    cdef IntArray nbrprocs              # List of neighboring processors
    cdef public double size             # total number of particles in this cell

    ############################################################################
    # Member functions
    ############################################################################
    # set the indices for the cell
    cpdef set_indices(self, int index, UIntArray lindices, UIntArray gindices)

    # compute the bounding box for a cell. Layers is used to determine
    # the factor times the cell size the bounding box is offset from
    # the cell.
    cdef _compute_bounding_box(self, double cell_size,
                               int layers)

cdef class NeighborCache:
    cdef int _dst_index
    cdef int _src_index
    cdef int _narrays
    cdef list _particles

    cdef int _n_threads
    cdef NNPS _nnps
    cdef UIntArray _pid_to_tid
    cdef UIntArray _start_stop
    cdef IntArray _cached
    cdef void **_neighbors
    # This is made public purely for testing!
    cdef public list _neighbor_arrays
    cdef int _last_avg_nbr_size

    cdef void get_neighbors_raw(self, size_t d_idx, UIntArray nbrs) nogil
    cpdef get_neighbors(self, int src_index, size_t d_idx, UIntArray nbrs)
    cpdef find_all_neighbors(self)
    cpdef update(self)

    cdef void _update_last_avg_nbr_size(self)
    cdef void _find_neighbors(self, long d_idx) nogil

cdef class NNPSBase:
    ##########################################################################
    # Data Attributes
    ##########################################################################
    cdef public list particles        # list of particle arrays
    cdef public list pa_wrappers      # list of particle array wrappers
    cdef public int narrays           # Number of particle arrays
    cdef bint use_cache               # Use cache or not.
    cdef public list cache            # The neighbor cache.
    cdef int src_index, dst_index     # The current source and dest indices

    cdef public DomainManager domain  # Domain manager
    cdef public bint is_periodic      # flag for periodicity

    cdef public int dim               # Dimensionality of the problem
    cdef public double cell_size      # Cell size for binning
    cdef public double hmin           # Minimum h
    cdef public double radius_scale   # Radius scale for kernel
    cdef IntArray cell_shifts         # cell shifts
    cdef public int n_cells           # number of cells

    # Testing function for brute force neighbor search. The return
    # list is of the same type of the local and global ids (uint)
    cpdef brute_force_neighbors(self, int src_index, int dst_index,
                                size_t d_idx, UIntArray nbrs)

    cpdef get_nearest_particles_no_cache(self, int src_index, int dst_index,
            size_t d_idx, UIntArray nbrs, bint prealloc)

    cdef void find_nearest_neighbors(self, size_t d_idx, UIntArray nbrs) nogil

    cpdef get_nearest_particles(self, int src_index, int dst_index,
                                size_t d_idx, UIntArray nbrs)
    cpdef set_context(self, int src_index, int dst_index)

# Nearest neighbor locator
cdef class NNPS(NNPSBase):
    ##########################################################################
    # Data Attributes
    ##########################################################################
    cdef public DoubleArray xmin      # co-ordinate min values
    cdef public DoubleArray xmax      # co-ordinate max values
    cdef public NeighborCache current_cache  # The current cache
    cdef public double _last_domain_size # last size of domain.

    cdef public bint sort_gids        # Sort neighbors by their gids.

    ##########################################################################
    # Member functions
    ##########################################################################
    # Main binning routine for NNPS for local particles. This clears
    # the current cell data, re-computes the cell size and bins all
    # particles locally.
    cpdef update(self)

    # Index particles given by a list of indices. The indices are
    # assumed to be of type unsigned int and local to the NNPS object
    cpdef _bin(self, int pa_index, UIntArray indices)

    cdef void _sort_neighbors(self, unsigned int* nbrs, size_t length,
                              unsigned int *gids) nogil

    # compute the min and max for the particle coordinates
    cdef _compute_bounds(self)

    cdef void find_nearest_neighbors(self, size_t d_idx, UIntArray nbrs) nogil

    cdef void get_nearest_neighbors(self, size_t d_idx,
                                      UIntArray nbrs) nogil

    # Neighbor query function. Returns the list of neighbors for a
    # requested particle. The returned list is assumed to be of type
    # unsigned int to follow the type of the local and global ids.
    # This method will never use the cached values.  If prealloc is set
    # to True it will assume that the neighbor array has enough space for
    # all the new neighbors and directly set the values in the array.
    cpdef get_nearest_particles_no_cache(self, int src_index, int dst_index,
                            size_t d_idx, UIntArray nbrs, bint prealloc)

    # Neighbor query function. Returns the list of neighbors for a
    # requested particle. The returned list is assumed to be of type
    # unsigned int to follow the type of the local and global ids.
    cpdef get_nearest_particles(self, int src_index, int dst_index,
                                size_t d_idx, UIntArray nbrs)

    cpdef get_spatially_ordered_indices(self, int pa_index, LongArray indices)

    cpdef set_context(self, int src_index, int dst_index)

    cpdef spatially_order_particles(self, int pa_index)

    # refresh any data structures needed for binning
    cpdef _refresh(self)
